/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    album_environment.cc
 * @brief   AlbumEnvironment class definition.
 * @date    2008
 *
 */

#include <string>

#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>

#include "album_environment.h"
#include "texture_generator.h"
#include "factories/factory_environment.h"

namespace
{
	// half of the width of the album environment
	const float AE_WIDTH_HALF = 200.0f;

	// half of the height of the album environment
	const float AE_HEIGHT_HALF = 150.0f;

	// number of vertices in the album environment
	const unsigned AE_BOX_SIZE = 8;

	// number of faces of the album environment
	const unsigned AE_BOX_FACES = 6;

	// number of vertices in an album environment (including a '-1' ending index)
	const unsigned AE_BOX_FACE_SIZE = 5;

	// number of vertices used for texture of a album environment face
	// (including a '-1' ending index)
	const unsigned AE_BOX_TEX_SIZE = 5;

	// position of the album environment vertices
	// (copied from http://www.root.cz/clanky/open-inventor-vesmirna-scena-3/)
	const float AE_BOX_VERTICES[AE_BOX_SIZE][3] =
	{
		{-AE_WIDTH_HALF, -AE_HEIGHT_HALF, -AE_WIDTH_HALF},  // front left bottom
		{ AE_WIDTH_HALF, -AE_HEIGHT_HALF, -AE_WIDTH_HALF},  // front right bottom
		{-AE_WIDTH_HALF,  AE_HEIGHT_HALF, -AE_WIDTH_HALF},  // front left top
		{ AE_WIDTH_HALF,  AE_HEIGHT_HALF, -AE_WIDTH_HALF},  // front right top
		{-AE_WIDTH_HALF, -AE_HEIGHT_HALF,  AE_WIDTH_HALF},  // back left bottom
		{ AE_WIDTH_HALF, -AE_HEIGHT_HALF,  AE_WIDTH_HALF},  // back right bottom
		{-AE_WIDTH_HALF,  AE_HEIGHT_HALF,  AE_WIDTH_HALF},  // back left top
		{ AE_WIDTH_HALF,  AE_HEIGHT_HALF,  AE_WIDTH_HALF}   // back right top
	};

	// indices into AE_BOX_VERTICES (-1 ends a TriangleStrip)
	const int32_t AE_BOX_INDICES[AE_BOX_FACES][AE_BOX_FACE_SIZE] =
	{
		{6, 7, 4, 5, -1},   // back face
		{2, 3, 0, 1, -1},   // front face
		{2, 6, 0, 4, -1},   // left face
		{7, 3, 5, 1, -1},   // right face
		{2, 3, 6, 7, -1},   // top face
		{4, 5, 0, 1, -1}    // bottom face
	};

	// indices of the texture for each face
	const int32_t AE_BOX_TEX_COORD_INDICES[AE_BOX_TEX_SIZE] =
	{
		0, 1, 2, 3, -1
	};
}


AlbumEnvironment::AlbumEnvironment()
{
	// initialize the root node
	_rootNode = new SoGroup();
	SoSeparator* environ = new SoSeparator();

	// set correct light model for the background
	SoLightModel* lightModel = new SoLightModel();
	lightModel->model.setValue(SoLightModel::PHONG);
	environ->addChild(lightModel);

	// add album vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, AE_BOX_SIZE, AE_BOX_VERTICES);
	environ->addChild(coords);

	SoTextureCoordinate2* coord = new SoTextureCoordinate2();
	coord->point.set1Value(0, SbVec2f(0.0f, 1.0f));
	coord->point.set1Value(1, SbVec2f(1.0f, 1.0f));
	coord->point.set1Value(2, SbVec2f(0.0f, 0.0f));
	coord->point.set1Value(3, SbVec2f(1.0f, 0.0f));
	environ->addChild(coord);

	AlbumContainer collection = TextureGenerator::generateAlbum("sweet");

	// array for space background textures
	for (unsigned i = 0; i < AE_BOX_FACES; ++i) {
		// add the album textures
		// TODO: dispose of magic constants & check for valid return value
		collection[i]->model.setValue(SoTexture2::DECAL);  // discard lighting result
		environ->addChild(collection[i]);

		// create the album walls
		SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
		strip->coordIndex.setValues(0, AE_BOX_FACE_SIZE, AE_BOX_INDICES[i]);
		strip->textureCoordIndex.setValues(0, AE_BOX_TEX_SIZE, AE_BOX_TEX_COORD_INDICES);
		environ->addChild(strip);
	}

	_rootNode->addChild(environ);


	// create lights
	SoPointLight* light = new SoPointLight();
	light->location.setValue(SbVec3f(-10.0f, -10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);

	light = new SoPointLight();
	light->location.setValue(SbVec3f(10.0f, 10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);
}

AlbumEnvironment::~AlbumEnvironment()
{
}

void AlbumEnvironment::timeUpdate(float /* dt */)
{
	// do nothing
}

// Functions and constants needed by Factories::FactoryEnvironment
namespace
{
	// SpaceEnvironment identifier for the environments factory
	const std::string ALBUM_ENVIRONMENT_ID = "Album";

	// Function that returns an instance of the AlbumEnvironment class
	Environment *getAlbumEnvironmentInstance()
	{
		return new AlbumEnvironment::AlbumEnvironment();
	}

	// Register the selected environment at the environments factory
	const bool ALBUM_ENVIRONMENT_REGISTERED = Factories::FactoryEnvironment::instance().
		registerObject(ALBUM_ENVIRONMENT_ID, getAlbumEnvironmentInstance);
}

