/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    avatar.h
 * @brief   Header file for the Avatar abstract class
 * @date    2008
 *
 * This file includes the declaration of the Avatar abstract class, the class
 * encapsulating player avatar.
 */

#ifndef _AVATAR_H_
#define _AVATAR_H_

#include "camera.h"
#include "time_updateable.h"
#include "graphical_node.h"
#include "templates/observer.h"
#include "templates/callable_back.h"

class PlayField;

/**
 * @brief   Abstract class encapsulating player avatar
 *
 * This abstract class encapsulates player avatar. The avatar has a graphical
 * representation in the OpenInventor. This class is a reference object
 * (cannot be copied).
 */
class Avatar: public TimeUpdateable, public GraphicalNode,
	public Templates::Observer<Camera>
{
	private:

		// Disable copy constructor and assignment operator
		Avatar(const Avatar &);
		Avatar & operator=(const Avatar &);

	protected:

		/**
		 * @brief  A static method that generates a model of the bubble hammer
		 *
		 * @return  A SoSeparator with the bubble hammer
		 */
		static SoSeparator* getBubbleHammer();

	public:

		/**
		 * @brief  Default constructor
		 *
		 * Default constructor of the Avatar class
		 */
		Avatar();

		/**
		 * @brief  Destructor
		 */
		virtual ~Avatar();

		/**
		 * @brief  The method that starts the avatar bubble-firing sequence
		 *
		 * This method starts the avatar bubble-firing sequence. When it is done,
		 * it calls the @a fireDone "action performed" method.
		 *
		 * @param[in]  fireDone  Object on which the "action performed" method
		 *                       should be called when the bubble-firing
		 *                       sequence is completed
		 */
		virtual void fire(Templates::CallableBack<Avatar> *fireDone) =0;
};

#endif
