/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    basic_play_box_graphics.cc
 * @brief   BasicPlayBoxGraphics class definition.
 * @date    2008
 *
 */

#include <string>

#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>

#include "basic_play_box_graphics.h"
#include "texture_generator.h"
#include "play_field.h"
#include "factories/factory_play_box_graphics.h"

namespace
{
	// The red component of the play box colour
	const float BOX_COLOUR_RED = 0.25f;

	// The green component of the play box colour
	const float BOX_COLOUR_GREEN = 0.25f;

	// The blue component of the play box colour
	const float BOX_COLOUR_BLUE = 0.85f;

	// The alpha component of the play box colour
	const float BOX_TRANSPARENCY = 0.9f;

	// Half of the side of the play box
	const float BOX_SIDE_HALF = PlayField::getSizeHalfSize();

	// Depth of the play box
	const float BOX_DEPTH = PlayField::getDepth();

	// number of vertices in the basic play field
	const unsigned BPF_SIZE = 8;

	// number of faces of the basic play field
	const unsigned BPF_FACES = 4;

	// number of vertices in a basic play field face
	// (including a '-1' ending index)
	const unsigned BPF_FACE_SIZE = 5;

	// number of lines in the play box
	const unsigned BPF_LINES = 4;

	// number of vertices used for texture of a basic play field face
	// (including a '-1' ending index)
	const unsigned BPF_LINE_SIZE = 3;

	// position of the basic play field vertices
	// (copied from http://www.root.cz/clanky/open-inventor-vesmirna-scena-3/)
	const float BPF_VERTICES[BPF_SIZE][3] =
	{
		{-BOX_SIDE_HALF, -BOX_SIDE_HALF,         0},  // front left bottom
		{ BOX_SIDE_HALF, -BOX_SIDE_HALF,         0},  // front right bottom
		{-BOX_SIDE_HALF,  BOX_SIDE_HALF,         0},  // front left top
		{ BOX_SIDE_HALF,  BOX_SIDE_HALF,         0},  // front right top
		{-BOX_SIDE_HALF, -BOX_SIDE_HALF, BOX_DEPTH},  // back left bottom
		{ BOX_SIDE_HALF, -BOX_SIDE_HALF, BOX_DEPTH},  // back right bottom
		{-BOX_SIDE_HALF,  BOX_SIDE_HALF, BOX_DEPTH},  // back left top
		{ BOX_SIDE_HALF,  BOX_SIDE_HALF, BOX_DEPTH}   // back right top
	};

	// indices into BPF_VERTICES (-1 ends a TriangleStrip)
	const int32_t BPF_INDICES[BPF_FACES][BPF_FACE_SIZE] =
	{
		{6, 4, 2, 0, -1},   // left face
		{5, 7, 1, 3, -1},   // right face
		{2, 3, 6, 7, -1},   // top face
		{4, 5, 0, 1, -1}    // bottom face
	};

	// indices of the back wall into BPF_VERTICES (-1 ends a TriangleStrip)
	const int32_t BPF_BWALL_INDICES[BPF_FACE_SIZE] =
	{
		0, 2, 1, 3, -1
	};

	// indices of line vertices (from BPF_VERTICS) for the basic play box
	// (-1 ends a LineStrip)
	const int32_t BPF_LINE_INDICES[BPF_LINES][BPF_LINE_SIZE] =
	{
		{0, 4, -1},
		{1, 5, -1},
		{2, 6, -1},
		{3, 7, -1}
	};
}


BasicPlayBoxGraphics::BasicPlayBoxGraphics()
{
	_rootNode = new SoSeparator();

	// create the walls of the box
	_rootNode->addChild(createWalls());

	// create the back wall of the box
	_rootNode->addChild(createBackWall());
}

SoGroup* BasicPlayBoxGraphics::createWalls()
{
	SoSeparator* sep = new SoSeparator();

	// add play field vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, BPF_SIZE, BPF_VERTICES);
	sep->addChild(coords);

	// the line contour of the box
	SoSeparator* boxLineContour = new SoSeparator();

	// set correct light model for the box
	SoLightModel* lightModel = new SoLightModel();
	lightModel->model.setValue(SoLightModel::BASE_COLOR);
	boxLineContour->addChild(lightModel);

	// render play field lines
	SoBaseColor* lineColour = new SoBaseColor();
	lineColour->rgb.setValue(1.f, 1.f, 1.f);
	boxLineContour->addChild(lineColour);

	// change the drawing style to draw lines
	SoDrawStyle* drawStyle = new SoDrawStyle();
	drawStyle->style.setValue(SoDrawStyle::LINES);
	drawStyle->lineWidth.setValue(2.0f);
	boxLineContour->addChild(drawStyle);

	// draw play field lines
	SoIndexedLineSet* lines = new SoIndexedLineSet();
	// send the whole array of indices to the line set
	lines->coordIndex.setValues(0, BPF_LINES*BPF_LINE_SIZE, *BPF_LINE_INDICES);
	boxLineContour->addChild(lines);

	sep->addChild(boxLineContour);

	// Material
	SoMaterial *material = new SoMaterial();
	material->diffuseColor.setValue(
		BOX_COLOUR_RED, BOX_COLOUR_GREEN, BOX_COLOUR_BLUE);
	material->transparency.setValue(BOX_TRANSPARENCY);
	sep->addChild(material);

	// create walls of the play field
	SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
	// send the whole field of indices to the triangle strip
	strip->coordIndex.setValues(0, BPF_FACES * BPF_FACE_SIZE, *BPF_INDICES);
	sep->addChild(strip);

	return sep;
}

SoGroup* BasicPlayBoxGraphics::createBackWall()
{
	// draw the back wall
	SoSeparator* backWallSep = new SoSeparator();

	// add play field vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, BPF_SIZE, BPF_VERTICES);
	backWallSep->addChild(coords);

	SoMaterial* material = new SoMaterial();
	//material->diffuseColor.setValue(155.0f/256.0f, 144.0f/256.0f, 24.0f/256.0f);
	material->diffuseColor.setValue(0.2f, 0.2f, 0.2f);
	backWallSep->addChild(material);

	backWallSep->addChild(_wallTranslation);

	// create the back wall of the play field
	SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
	// send the whole field of indices to the triangle strip
	strip->coordIndex.setValues(0, BPF_FACE_SIZE, BPF_BWALL_INDICES);
	backWallSep->addChild(strip);

	return backWallSep;
}

BasicPlayBoxGraphics::~BasicPlayBoxGraphics()
{
}

void BasicPlayBoxGraphics::timeUpdate(float /*dt*/)
{
	// do nothing
}

// Functions and constants needed by Factories::FactoryPlayBoxGraphics
namespace
{
	// BasicPlayBoxGraphics identifier for the play box graphics factory
	const std::string BASIC_PLAY_BOX_GRAPHICS_ID = "Basic";

	// Function that returns an instance of the BasicPlayBoxGraphics class
	PlayBoxGraphics *getBasicPlayBoxGraphicsInstance()
	{
		return new BasicPlayBoxGraphics::BasicPlayBoxGraphics();
	}

	// Register the selected play box graphics at the play box graphics factory
	const bool BASIC_PLAY_BOX_GRAPHICS_REGISTERED = Factories::FactoryPlayBoxGraphics::instance().
		registerObject(BASIC_PLAY_BOX_GRAPHICS_ID, getBasicPlayBoxGraphicsInstance);
}
