/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    bubble.cc
 * @brief   Bubble class
 * @date    2008
 *
 * Implementation of the bubble class
 *
 */

#include <cassert>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoComplexity.h>

#include "bubble.h"
#include "explosion.h"
#include "play_field.h"

const float Bubble::TRANSPARENCY_TRANSPARENT = 0.5f;
const float Bubble::TRANSPARENCY_SOLID = 0.0f;

int Bubble::freeTypeId = 0;
float Bubble::speed = 25.0f;

Bubble::Bubble(int id, const SbColor& color, Explosion* explosion):
	_typeId(id), _motionVector(0.0f, 0.0f, 0.0f), _position(0),
	_color(color), _explosion(explosion), _explodeDone(0), _material(0)
{
	createOIGraph();
}

void Bubble::createOIGraph()
{
	_rootNode = new SoSeparator();

	// complexity of a bubble
	SoComplexity* earthComplexity = new SoComplexity();
	earthComplexity->value.setValue(2.0f);
	_rootNode->addChild(earthComplexity);

	// Create a material of the bubble
	_material = new SoMaterial();
	_material->diffuseColor.setValue(_color);
	_material->specularColor.setValue(SbColor(1.0f, 1.0f, 1.0f));
	_material->shininess.setValue(0.5f);
	_material->transparency.setValue(TRANSPARENCY_TRANSPARENT);
	_rootNode->addChild(_material);

	// Create a transition object
	_position = new SoTranslation();
	_rootNode->addChild(_position);

	// The bubble
	SoSphere* bubble = new SoSphere();
	bubble->radius.setValue(getRadius());
	_rootNode->addChild(bubble);

//	// draw the wireframe around the bubble
//	SoDrawStyle *drawStyle = new SoDrawStyle;
//	drawStyle->style.setValue(SoDrawStyle::LINES);
//	drawStyle->lineWidth.setValue(3.f);
//	_rootNode->addChild(drawStyle);
//
//	bubble = new SoSphere();
//	bubble->radius.setValue(getRadius());
//	_rootNode->addChild(bubble);
}

Bubble::Bubble(const SbColor& color, Explosion* explosion):
	_typeId(freeTypeId), _motionVector(0.0f, 0.0f, 0.0f), _position(0),
	_color(color), _explosion(explosion), _explodeDone(0), _material(0)
{
	++Bubble::freeTypeId;
	createOIGraph();
}

Bubble::~Bubble()
{
	delete _explosion;
	_explosion = 0;
}

void Bubble::timeUpdate(float dt)
{
	// If the bubble is still flying, move it
	if (_motionVector != SbVec3f(0.0f, 0.0f, 0.0f)) {
		// Check if the motion vector has unit length
		assert((0.99f < _motionVector.length()) &&
				(_motionVector.length() < 1.01f));
		// move the bubble
		_position->translation.setValue((dt * _motionVector) +
				_position->translation.getValue());
	}

	_explosion->timeUpdate(dt);
}

void Bubble::callableBackNotify(Explosion* /*subject*/)
{
	// Bubble has exploded
	_explodeDone->callableBackNotify(this);
}

Bubble* Bubble::clone() const
{
	return new Bubble(_typeId, _color, _explosion->clone());
}

void Bubble::setMotionVector(const SbVec3f& newMotionVector)
{
	_motionVector = newMotionVector;
	_motionVector.normalize();
}

bool Bubble::collidesWithBubble(const SbVec3f& other) const
{
	float distance = (getPosition() - other).length();
	return distance <= (2*Bubble::getRadius());
}

void Bubble::enableTransparency()
{
	_material->transparency.setValue(TRANSPARENCY_TRANSPARENT);
}

void Bubble::disableTransparency()
{
	_material->transparency.setValue(TRANSPARENCY_SOLID);
}

void Bubble::explode(Templates::CallableBack<Bubble>* explodeDone)
{
	_explodeDone = explodeDone;

	_explosion->setPosition(_position->translation.getValue());
	_explosion->explode(this);
	_rootNode->removeAllChildren();
	_rootNode->addChild(_explosion->getRootNode());
}
