/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    bubble.h
 * @brief   Bubble class
 * @date    2008
 *
 * Declaration of the bubble class
 *
 */

#ifndef _BUBBLE_H_
#define _BUBBLE_H_

#include <Inventor/SbColor.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoMaterial.h>

#include "graphical_node.h"
#include "time_updateable.h"
#include "play_field.h"
#include "bubble_field.h"
#include "templates/callable_back.h"

class Explosion;

/**
 * @brief   Bubble in the game
 *
 * Bubble can have its own color and can be static or moving
 */
class Bubble: public TimeUpdateable, public GraphicalNode,
	public Templates::CallableBack<Explosion>
{
	private:

		/// Transparency value of a transparent bubble
		static const float TRANSPARENCY_TRANSPARENT;

		/// Transparency value of a solid bubble
		static const float TRANSPARENCY_SOLID;

		/// Stores first free number than identifies types of bubble
		static int freeTypeId;

		/// Basic speed of the bubble
		static float speed;

		/// Id that identifies the type of the bubble
		int _typeId;

		/// Motion vector (unit vector)
		SbVec3f _motionVector;

		/// Position (point is represented by vector)
		SoTranslation* _position;

		/// Diffuse color of the bubble
		SbColor _color;

		/// Explosion effect
		Explosion* _explosion;

		/// Object on which the "action performed" method should be called when
		/// the explosion is completed
		Templates::CallableBack<Bubble>* _explodeDone;

		/// The material of the bubble
		SoMaterial* _material;

	private:

		/**
		 * @brief  Private constructor
		 *
		 * @param[in]  id         Identifies type of the bubble
		 * @param[in]  color      Color of the bubble
		 * @param[in]  explosion  Explosion effect
		 */
		Bubble(int id, const SbColor& color, Explosion* explosion);

		/**
		 * @brief Creates OI objects and add them to the SoSeparator
		 *
		 * This method can be called only once (in the constructor)
		 *
		 */
		void createOIGraph();

	public:

		/**
		 * @brief  Public constructor
		 *
		 * @param[in]  color      Color of the bubble
		 * @param[in]  explosion  Explosion effect
		 *
		 * Creates new type of bubble. You should use clone() in most cases instead
		 */
		Bubble(const SbColor& color, Explosion* explosion);

		/**
		 * @brief  Destructor
		 */
		~Bubble();

		// Override
		virtual void timeUpdate(float dt);

		// Override
		virtual void callableBackNotify(Explosion* subject);

		/**
		 * @brief  Clones this bubble
		 *
		 * @return Returns pointer to a new alloced exact copy of itself
		 */
		Bubble* clone() const;

		/**
		 * @brief Sets motion vector
		 *
		 * Given vector is transformed to unit vector before the assignment
		 *
		 * @param[in] newMotionVector New motion vector
		 */
		void setMotionVector(const SbVec3f& newMotionVector);

		/**
		 * @brief Gets motion vector
		 *
		 * @returns Motion vector
		 *
		 */
		const SbVec3f& getMotionVector() const
		{
			return _motionVector;
		}

		/**
		 * @brief Sets new position
		 *
		 * @param[in]  newPosition  Vector which determines new position of the bubble
		 */
		void setPosition(const SbVec3f& newPosition)
		{
			_position->translation = newPosition;
		}

		/**
		 * @brief Gets position
		 *
		 * @returns Position of the bubble
		 */
		const SbVec3f& getPosition() const
		{
			return _position->translation.getValue();
		}

		/**
		 * @brief Gets speed of the bubbles
		 *
		 * @returns Speed of the bubbles
		 */
		static float getSpeed() { return speed; }

		/**
		 * @brief Gets radius of all bubbles
		 *
		 * @returns Radius of the bubbles
		 */
		static float getRadius()
		{
			return PlayField::getSizeHalfSize() / BubbleField::getNumberOfBubblesInRow();
		}

		/**
		 * @brief Returns type of the bubble
		 */
		int getType() { return _typeId; }

		/**
		 * @brief Returns true if the Bubble collides with other Bubble
		 *
		 * @param[in]  other  Position of other bubble
		 */
		bool collidesWithBubble(const SbVec3f& other) const;

		/**
		 * @brief  Makes the bubble transparent
		 */
		virtual void enableTransparency();

		/**
		 * @brief  Makes the bubble solid
		 */
		virtual void disableTransparency();

		/**
		 * @brief Explodes the bubble
		 *
		 * @param[in]  explodeDone  Object on which the "action performed" method
		 *                          should be called when the bubble has exploded
		 */
		void explode(Templates::CallableBack<Bubble>* explodeDone);
};

#endif
