/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    bubble_field.cc
 * @brief   BubbleField class definition.
 * @date    2008
 *
 */

#include <cstdlib>
#include <cassert>
#include <cmath>
#include <list>
#include <set>
#include <vector>
#include <algorithm>

#include <Inventor/SbLinear.h>

#include "bubble.h"
#include "bubble_field.h"
#include "play_field.h"

const unsigned BubbleField::BUBBLES_IN_ROW = 4;

namespace
{
	// Half of the side of the play box
	const float BOX_SIDE_HALF = PlayField::getSizeHalfSize();

	// Depth of the playbox
	const float BOX_DEPTH = PlayField::getDepth();

	// Distance between two plates of bubbles
	const float PLATE_DISTANCE = std::sqrt(3.0f) * Bubble::getRadius();

	// Diameter of a bubble
	const float BUBBLE_DIAMETER = 2 * Bubble::getRadius();

	// Number of bubbles in depth
	const unsigned BUBBLES_IN_DEPTH = static_cast<unsigned>(std::floor(
			(BOX_DEPTH - Bubble::getRadius()) / PLATE_DISTANCE) + 1.0f);

	// Minimal number of bubbles in chain to destroy the bubbles
	const unsigned MIN_BUBBLES_TO_EXPLODE = 3;

	// OI coordinate of the bottom of the field (when bubble is under the ground)
	const float OI_BOTTOM = -((2 * BubbleField::getNumberOfBubblesInRow() + 1) * Bubble::getRadius());

	// Speed of falling bubble
	const float FALLING_BUBBLE_SPEED = 3;
}

/******************************************************************************/
// FallingBubblesStorage class

BubbleField::FallingBubblesStorage::FallingBubblesStorage():
	_storage()
{
	_rootNode = new SoGroup();
}

BubbleField::FallingBubblesStorage::~FallingBubblesStorage()
{
	// Free memory of stored bubbles
	for (StorageType::iterator it = _storage.begin();
			it != _storage.end(); ++it) {
		Bubble *bubble = *it;
		if (bubble != 0) {
			_rootNode->removeChild(bubble->getRootNode());
			delete bubble;
			*it = 0;
		}
	}
}

void BubbleField::FallingBubblesStorage::timeUpdate(float dt)
{
	// Call timeUpdate of all stored bubbles
	StorageType::iterator it = _storage.begin();
	while (it != _storage.end()) {
		Bubble* bubble = *it;
		bubble->timeUpdate(FALLING_BUBBLE_SPEED * dt);
		const SbVec3f& position = bubble->getPosition();
		if (position[1] <= OI_BOTTOM) {
			_rootNode->removeChild(bubble->getRootNode());
			it = _storage.erase(it);
			delete bubble;
		} else {
			++it;
		}
	}
}

void BubbleField::FallingBubblesStorage::addBubble(Bubble* bubble)
{
	assert(bubble != 0);

	bubble->setMotionVector(SbVec3f(0.0f, -1.0f, 0.0f));
	bubble->enableTransparency();
	_storage.push_back(bubble);
	_rootNode->addChild(bubble->getRootNode());
}

/******************************************************************************/
// BubbleField class

unsigned BubbleField::bubblesInPlateRow(unsigned index) const
{
	return (index % 2 == 0) ? BUBBLES_IN_ROW : BUBBLES_IN_ROW - 1;
}

bool BubbleField::isValidPosition(unsigned plate, unsigned x, unsigned y) const
{
	const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
	return ((plate < _content.size()) && (x < ROW_WIDTH) && (y < ROW_WIDTH));
}

bool BubbleField::isInCollisionWithWall(const SbVec3f& position)
{
	return (position[2] < Bubble::getRadius());
}

bool BubbleField::bubbleCollidesInPlate(unsigned plate, const SbVec3f& position) const
{
	// For bubbles not in allowed space return false
	if (plate >= _content.size()) {
		return false;
	}

	const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
	for (unsigned i = 0; i < ROW_WIDTH; ++i) {
		for (unsigned j = 0; j < ROW_WIDTH; ++j) {
			if (_content[plate][i][j] != 0) {
				if (_content[plate][i][j]->collidesWithBubble(position)) {
					return true;
				}
			}
		}
	}

	return false;
}

void BubbleField::addCollidingBubblesInPlateToVector(unsigned plate, const Bubble* bubble,
		std::vector<const Bubble*>& collidingBubbles) const
{
	// For bubbles not in allowed space don't add anything
	if (plate >= _content.size()) {
		return;
	}
	const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
	for (unsigned i = 0; i < ROW_WIDTH; ++i) {
		for (unsigned j = 0; j < ROW_WIDTH; ++j) {
			if (_content[plate][i][j] != 0) {
				if (_content[plate][i][j]->collidesWithBubble(bubble->getPosition())) {
					collidingBubbles.push_back(bubble);
				}
			}
		}
	}
}

void BubbleField::solveCollisionWithWall(Bubble* bubble)
{
	SbVec3f position = bubble->getPosition();

	unsigned x = oiXYCoordinateToLocal(position[0], 0);
	unsigned y = oiXYCoordinateToLocal(position[1], 0);

	// Z coordinate of the bubble at the wall is always 0
	assert(_content[0][x][y] == 0);
	_content[0][x][y] = bubble;
	bubble->setPosition(BubbleField::localCoordinatesToOI(0, x, y));
	bubble->setMotionVector(SbVec3f(0.0f, 0.0f, 0.0f));
	destroyBubblesInCluster(Coordinates(0, x, y));
}

void BubbleField::solveCollisionWithBubble(Bubble* bubble)
{
	// Search for bubbles that are in collision
	assert(getCollidingBubbles(bubble).size() > 0);

	// Search for free place and add bubble there
	BubbleField::Coordinates coords = searchForPosition(bubble);
	_content[coords.plate][coords.x][coords.y] = bubble;
	bubble->setPosition(localCoordinatesToOI(coords));
	bubble->setMotionVector(SbVec3f(0.0f, 0.0f, 0.0f));

	// Search for bubbles of same type in one cluster
	destroyBubblesInCluster(coords);
}

void BubbleField::destroyBubblesInCluster(const Coordinates& position)
{
	assert(_content[position.plate][position.x][position.y] != 0);

	// Initialization
	std::set<Coordinates> open; // bubbles to be processed
	std::set<Coordinates> closed; // processed or being in open
	std::set<Coordinates> found; // found bubbles in cluster
	int type = _content[position.plate][position.x][position.y]->getType();
	open.insert(position);
	closed.insert(position);
	found.insert(position);

	// For each item in open search for neighbours and find out their type
	while (!open.empty()) {
		// Pop new coordinates
		Coordinates coords = *(open.begin());
		open.erase(open.begin());
		// Search for neighbours
		std::list<Coordinates> neighbours = getNeighboursCoords(coords);
		for (std::list<Coordinates>::iterator it = neighbours.begin();
				it != neighbours.end(); ++it) {
			Coordinates looked = *it;
			Bubble* lookedBubble = _content[looked.plate][looked.x][looked.y];
			if ((lookedBubble != 0) && (lookedBubble->getType() == type)) {
				// Same type bubble found
				if (closed.find(looked) == closed.end()) {
					// It hasn't been processed yet
					found.insert(looked);
					open.insert(looked);
					closed.insert(looked);
				}
			}
		}
	}

	// Check if the bubbles should explode
	if (found.size() >= MIN_BUBBLES_TO_EXPLODE) {
		// Explosion is comming
		_explodingBubblesCount += found.size();
		for (std::set<Coordinates>::iterator it = found.begin();
				it != found.end(); ++it) {
			Coordinates coord = *it;
			Bubble* bubble = _content[coord.plate][coord.x][coord.y];
			assert(bubble != 0);
			bubble->explode(this);
			_explodingBubbles.push_back(bubble);
			_content[coord.plate][coord.x][coord.y] = 0;
		}

		// Look for bubbles not connected with the rear wall
		removeBubblesNotConnectedWithWall();
		// Look if the game is finished
		for (unsigned x = 0; x < bubblesInPlateRow(0); ++x) {
			for (unsigned y = 0; y < bubblesInPlateRow(0); ++y) {
				if (_content[0][x][y] != 0) {
					return;
				}
			}
		}

		// This place is reached only when first plate is empty and
		// it means that the user won the game
		if (_successGameEndFunc != 0) {
			_successGameEndFunc();
		}
	} else {
		// No explosion, adding of the bubble is done
		_addBubbleDone->callableBackNotify(this);
	}
}

void BubbleField::removeBubblesNotConnectedWithWall()
{
	// Initialization
	std::set<Coordinates> open; // bubbles to be processed
	std::set<Coordinates> closed; // processed or being in open
	for (unsigned x = 0; x < BUBBLES_IN_ROW; ++x) {
		for (unsigned y = 0; y < BUBBLES_IN_ROW; ++y) {
			if (_content[0][x][y] != 0) {
				Coordinates position = Coordinates(0, x, y);
				open.insert(position);
				closed.insert(position);
			}
		}
	}

	// For each item in open search for neighbours and add them to open and closed
	while (!open.empty()) {
		// Pop new coordinates
		Coordinates coords = *(open.begin());
		open.erase(open.begin());
		// Search for neighbours
		std::list<Coordinates> neighbours = getNeighboursCoords(coords);
		for (std::list<Coordinates>::iterator it = neighbours.begin();
				it != neighbours.end(); ++it) {
			Coordinates looked = *it;
			if ((_content[looked.plate][looked.x][looked.y]) &&
					(closed.find(looked) == closed.end())) {
				// We have found not processed position
				closed.insert(looked);
				open.insert(looked);
			}
		}
	}

	// Search for positions with bubble not in closed list, they are not connected
	// to the wall
	for (unsigned plate = 0; plate < BUBBLES_IN_DEPTH; ++plate) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
		for (unsigned x = 0; x < ROW_WIDTH; ++x) {
			for (unsigned y = 0; y < ROW_WIDTH; ++y) {
				Bubble* bubble = _content[plate][x][y];
				if ((bubble != 0) && (closed.find(Coordinates(plate, x, y)) == closed.end())) {
					// Unconected bubble found
					_fallingBubbles.addBubble(bubble);
					_content[plate][x][y] = 0;
					_rootNode->removeChild(bubble->getRootNode());
				}
			}
		}
	}
}

BubbleField::Coordinates BubbleField::searchForPosition(Bubble* bubble) const
{
	SbVec3f bubblePosition = bubble->getPosition();
	// This constant has to be created because C++ doesn't have useful for cycle
	// (for each)
	const unsigned INTERVAL_BOUNDARIES = 2;
	// Get neighbour plate indexes
	const unsigned plates[INTERVAL_BOUNDARIES] = {
		oiZCoordinateToLocal(bubblePosition[2], std::ceil),
		oiZCoordinateToLocal(bubblePosition[2], std::floor)};
	// Initialize variables for next algorithm
	float minDistance = INFINITY; // minimal distance to a bubble
	BubbleField::Coordinates coord(0, 0, 0); // coords where to place the given bubble
	// Iterate through neighbour plate indexes
	for (unsigned plate = 0; plate < INTERVAL_BOUNDARIES; ++plate) {
		// Get neighbour indexes of x and y for given plate
		const unsigned xs[INTERVAL_BOUNDARIES] = {
			oiXYCoordinateToLocal(bubblePosition[0], plates[plate], std::ceil),
			oiXYCoordinateToLocal(bubblePosition[0], plates[plate], std::floor)};
		const unsigned ys[INTERVAL_BOUNDARIES] = {
			oiXYCoordinateToLocal(bubblePosition[1], plates[plate], std::ceil),
			oiXYCoordinateToLocal(bubblePosition[1], plates[plate], std::floor)};
		for (unsigned x = 0; x < INTERVAL_BOUNDARIES; ++x) {
			for (unsigned y = 0; y < INTERVAL_BOUNDARIES; ++y) {
				// Coordinates of the found bubble position (current one)
				BubbleField::Coordinates found(plates[plate], xs[x], ys[y]);
				if (!isValidPosition(found.plate, found.x, found.y)) {
					continue; // Position is not in the BubbleField
				}
				// Distance from the found bubble position to the given bubble
				float distance = (localCoordinatesToOI(found) - bubblePosition).length();
				// Replace found position only if the current place is closer than the
				// old one, it is free and connected to a bubble connected to a wall
				if ((distance < minDistance) && (_content[found.plate][found.x][found.y] == 0) &&
						isConnectedWithWall(found)) {
					// New position is found
					minDistance = distance;
					coord = found;
				}
			}
		}
	}
	assert(minDistance < INFINITY); // Assert that we have found something

	return coord;
}

std::vector<const Bubble*> BubbleField::getCollidingBubbles(Bubble* bubble) const
{
	std::vector<const Bubble*> collidingBubbles;

	SbVec3f position = bubble->getPosition();
	unsigned zFg = oiZCoordinateToLocal(position[2], std::ceil);
	unsigned zBg = oiZCoordinateToLocal(position[2], std::floor);;

	addCollidingBubblesInPlateToVector(zFg, bubble, collidingBubbles);
	addCollidingBubblesInPlateToVector(zBg, bubble, collidingBubbles);

	return collidingBubbles;
}

std::list<BubbleField::Coordinates> BubbleField::getNeighboursCoords(
		const BubbleField::Coordinates& coords) const
{
	// Initialise list of potential neighbours to the ones in current plate
	std::list<BubbleField::Coordinates>potentialNeighbours;
	potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate, coords.x, coords.y - 1));
	potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate, coords.x, coords.y + 1));
	potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate, coords.x + 1, coords.y));
	potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate, coords.x - 1, coords.y));
	// Even plates: position of neighbour in odd plates is equal or less by one
	// Odd plates: position of neighbour in even plates is equal or higher by one
	const int OFFSET_LEN = 2;
	const int offsets[OFFSET_LEN] = {0, ((coords.plate % 2) == 0) ? -1 : 1};
	for (int i = 0; i < OFFSET_LEN; ++i) {
		for (int j = 0; j < OFFSET_LEN; ++j) {
			// Add potential neigbours in plates next to the current one
			unsigned x = coords.x + offsets[i];
			unsigned y = coords.y + offsets[j];
			potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate - 1, x, y));
			potentialNeighbours.push_back(BubbleField::Coordinates(coords.plate + 1, x, y));
		}
	}

	// Remove invalid positions
	potentialNeighbours.remove_if(isInvalidPosition(*this));

	return potentialNeighbours;
}

void BubbleField::addBubble(Bubble* bubble, unsigned plate, unsigned x, unsigned y)
{
	assert(isValidPosition(plate, x, y));
 	assert(_content[plate][x][y] == 0);

	// Add the bubble to the filed
	_content[plate][x][y] = bubble;

	// Add the bubble to the OI graph
	bubble->setPosition(BubbleField::localCoordinatesToOI(plate, x, y));
	bubble->disableTransparency();
	_rootNode->addChild(bubble->getRootNode());
}

void BubbleField::addNeighboursTowardsWall(const BubbleGenerator* bubbleGen,
	const Coordinates& coords)
{
	if (coords.plate == 0) {
		// We hit the wall
		return;
	}

	// Create coordinates for the new bubble
	unsigned newPlateCoord = coords.plate - 1;
	unsigned newXCoord = (coords.x == (bubblesInPlateRow(coords.plate) - 1) &&
		(coords.plate % 2) == 0) ? coords.x - 1 : coords.x;
	unsigned newYCoord = (coords.y == (bubblesInPlateRow(coords.plate) - 1) &&
		(coords.plate % 2) == 0) ? coords.y - 1 : coords.y;

	// If there is no bubble yet on the new coordinates, add a new bubble there
	if (_content[newPlateCoord][newXCoord][newYCoord] == 0) {
		addBubble(bubbleGen->generateBubble(), newPlateCoord, newXCoord, newYCoord);
	}

	// Check whether we're now connected to the wall
	Coordinates newCoords(newPlateCoord, newXCoord, newYCoord);
	Coordinates nearestToWallCoords(newCoords);
	if (!isConnectedWithWall(newCoords, std::vector<Coordinates>(), nearestToWallCoords)) {
		// No, we're still not, so continue
		addNeighboursTowardsWall(bubbleGen, nearestToWallCoords);
	}
}

bool BubbleField::isConnectedWithWall(const Coordinates& coords) const
{
	// Main function needs to have a reference to coordinates so we have to
	// create them even though we don't need them
	BubbleField::Coordinates unused(0, 0, 0);
	return isConnectedWithWall(coords, std::vector<Coordinates>(), unused);
}

bool BubbleField::isConnectedWithWall(const Coordinates& coords,
	std::vector<Coordinates> processedCoords, Coordinates &nearestToWallCoords) const
{
	if (coords.plate == 0) {
		// It is directly connected to the rear wall
		nearestToWallCoords = coords;
		return true;
	}

	// Recursively check whether some existing neighbour is connected to the wall
	std::list<Coordinates> neighbours = getNeighboursCoords(coords);
	for (std::list<Coordinates>::iterator iter = neighbours.begin();
			iter != neighbours.end(); ++iter) {
		// Is there a bubble on the selected position?
		if (_content[iter->plate][iter->x][iter->y] == 0) {
			continue;
		}

		// Check if we haven't already processed this neighbour
		if (std::find(processedCoords.begin(), processedCoords.end(),
				*iter) != processedCoords.end()) {
			// We have, so break the for loop
			continue;
		}

		// Recursively call isConnectedWithWall() to this neighbour
		std::vector<Coordinates> updatedProcessedCoords(processedCoords);
		updatedProcessedCoords.push_back(coords);
		Coordinates updatedNearestToWallCoords(*iter);
		if (isConnectedWithWall(*iter, updatedProcessedCoords, updatedNearestToWallCoords)) {
			nearestToWallCoords = updatedNearestToWallCoords;
			return true;
		}
	}

	if (coords.plate < nearestToWallCoords.plate) {
		nearestToWallCoords = coords;
	}
	return false;
}

unsigned BubbleField::oiXYCoordinateToLocal(float oiCoord, unsigned plate, float (*roundFunc)(float))
{
	float offset = ((plate % 2) == 0) ? Bubble::getRadius() : 2 * Bubble::getRadius();
	float normalizedCoord = (oiCoord + BOX_SIDE_HALF - offset);
	return static_cast<unsigned>(roundFunc(normalizedCoord / BUBBLE_DIAMETER));
}

unsigned BubbleField::oiZCoordinateToLocal(float oiCoord, float (*roundFunc)(float))
{
	float z = (oiCoord - Bubble::getRadius()) / PLATE_DISTANCE;
	return static_cast<unsigned>(roundFunc(z));
}

SbVec3f BubbleField::localCoordinatesToOI(unsigned plate, unsigned x, unsigned y)
{
	// There are more bubbles in an even plate row than in an odd plate row,
	// so to compensate it, bubble radius must be multiplied by two in odd plates
	const unsigned XY_RADIUS_MULT_COEF = (plate % 2 == 0) ? 1 : 2;

	return SbVec3f(-BOX_SIDE_HALF + XY_RADIUS_MULT_COEF * Bubble::getRadius() + x * BUBBLE_DIAMETER,
			-BOX_SIDE_HALF + XY_RADIUS_MULT_COEF * Bubble::getRadius() + y * BUBBLE_DIAMETER,
			plate * PLATE_DISTANCE + Bubble::getRadius());
}

SbVec3f BubbleField::localCoordinatesToOI(BubbleField::Coordinates local)
{
	return localCoordinatesToOI(local.plate, local.x, local.y);
}

BubbleField::BubbleField():
	_content(), _addBubbleDone(0), _explodingBubbles(),
	_explodingBubblesCount(0), _fallingBubbles(), _successGameEndFunc(0)
{
	_rootNode = new SoSeparator();
	_rootNode->addChild(_fallingBubbles.getRootNode());

	// Create empty content
	for (unsigned i = 0; i < BUBBLES_IN_DEPTH; ++i) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(i);
		BubblePlate plate = new Bubble** [ROW_WIDTH];
		for (unsigned j = 0; j < ROW_WIDTH; ++j) {
			plate[j] = new Bubble* [ROW_WIDTH];
			for (unsigned k = 0; k < ROW_WIDTH; ++k) {
				plate[j][k] = 0;
			}
		}
		_content.push_back(plate);
	}
}

BubbleField::~BubbleField()
{
	// Delete all remaining bubbles in the field (they were allocated
	// on the heap, so we have to delete them manually).
	// Also, delete correctly each plate.
	for (unsigned plate = 0; plate < _content.size(); ++plate) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
		for (unsigned i = 0; i < ROW_WIDTH; ++i) {
			for (unsigned j = 0; j < ROW_WIDTH; ++j) {
				Bubble *bubble = _content[plate][i][j];
				if (bubble != 0) {
					_rootNode->removeChild(bubble->getRootNode());
					delete bubble;
					_content[plate][i][j] = 0;
				}
			}

			delete[] _content[plate][i];
			_content[plate][i] = 0;
		}

		delete[] _content[plate];
		_content[plate] = 0;
	}
}

void BubbleField::setOnSuccessfullGameEndCallback(void (*callback)())
{
	_successGameEndFunc = callback;
}

void BubbleField::generateStartingLayout(const BubbleGenerator* bubbleGen,
	unsigned numOfInitiallyFilledPlates, float probCellContainBubble)
{
	// At first, just generate bubbles with the given probability to
	// the given number of plates (possible "problems" will be solved later)
	for (unsigned plate = 0; plate < _content.size() &&
			plate < numOfInitiallyFilledPlates; ++plate) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
		for (unsigned i = 0; i < ROW_WIDTH; ++i) {
			for (unsigned j = 0; j < ROW_WIDTH; ++j) {
				// There should be no bubbles at this position
				assert(_content[plate][i][j] == 0);

				if ((std::rand() / (RAND_MAX + 1.0)) <= probCellContainBubble) {
					addBubble(bubbleGen->generateBubble(), plate, i, j);
				}
			}
		}
	}

	// Now, check that there is no cluster of bubbles that is not connected
	// to the rear wall and if there is such a cluster, connect it to the wall
	// so the bubbles in the cluster won't fall down when the game starts
	for (unsigned plate = 0; plate < _content.size() &&
			plate < numOfInitiallyFilledPlates; ++plate) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
		for (unsigned i = 0; i < ROW_WIDTH; ++i) {
			for (unsigned j = 0; j < ROW_WIDTH; ++j) {
				if (_content[plate][i][j] != 0) {
					Coordinates bubbleCoords(plate, i, j);
					Coordinates nearestToWallCoords(plate, i, j);
					if (!isConnectedWithWall(bubbleCoords,
							std::vector<Coordinates>(), nearestToWallCoords)) {
						// This bubble is not connected to the wall, so connect
						// the neighbour that is the nearest one to the wall
						addNeighboursTowardsWall(bubbleGen, nearestToWallCoords);
					}
				}
			}
		}
	}
}

void BubbleField::addBubble(Bubble* bubble, Templates::CallableBack<BubbleField>* addBubbleDone)
{
	_addBubbleDone = addBubbleDone;
	_rootNode->addChild(bubble->getRootNode());

	if (isInCollisionWithWall(bubble->getPosition())) {
		solveCollisionWithWall(bubble);
	} else {
		solveCollisionWithBubble(bubble);
	}
}

bool BubbleField::isInCollision(const SbVec3f& position)
{
	if (isInCollisionWithWall(position)) {
		return true;
	} else {
		unsigned zFg = oiZCoordinateToLocal(position[2], std::ceil);
		unsigned zBg = oiZCoordinateToLocal(position[2], std::floor);;
		if (bubbleCollidesInPlate(zFg, position) ||
				bubbleCollidesInPlate(zBg, position)) {
			return true;
		}
		return false;
	}
}

void BubbleField::timeUpdate(float dt)
{
	// Forward the timeUpdate call to all stored bubbles
	for (unsigned plate = 0; plate < _content.size(); ++plate) {
		const unsigned ROW_WIDTH = bubblesInPlateRow(plate);
		for (unsigned i = 0; i < ROW_WIDTH; ++i) {
			for (unsigned j = 0; j < ROW_WIDTH; ++j) {
				if (_content[plate][i][j] != 0) {
					_content[plate][i][j]->timeUpdate(dt);
				}
			}
		}
	}

	// Forward the timeUpdate call to all exploding bubbles
	std::list<Bubble *>::iterator iter = _explodingBubbles.begin();
	while (iter != _explodingBubbles.end()) {
		if (*iter != 0) {
			(*iter)->timeUpdate(dt);
			++iter;
		} else {
			// Correctly remove the item from the container
			// (bubble was deleted in a callableBackNotify() call)
			iter = _explodingBubbles.erase(iter);
		}
	}

	// Forward the timeUpdate call to all falling bubbles
	_fallingBubbles.timeUpdate(dt);
}

void BubbleField::callableBackNotify(Bubble *subject)
{
	// Remove the bubble from the OI graph
	_rootNode->removeChild(subject->getRootNode());

	// Delete the memory allocated for the bubble
	// (item will be removed from the container in a timeUpdate() call)
	for (std::list<Bubble *>::iterator iter = _explodingBubbles.begin();
			iter != _explodingBubbles.end(); ++iter) {
		if (*iter == subject) {
			delete subject;
			subject = 0;
			*iter = 0;
			break;
		}
	}

	// Have all bubbles exploded?
	--_explodingBubblesCount;
	if (_explodingBubblesCount <= 0) {
		// All bubbles have exploded, so the addBubble() method finishes
		_addBubbleDone->callableBackNotify(this);
	}
}
