/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    bubble_field.h
 * @brief   BubbleField class declaration.
 * @date    2008
 *
 */

#ifndef _BUBBLE_FIELD_H_
#define _BUBBLE_FIELD_H_

#include <cmath>
#include <list>
#include <vector>

#include "bubble.h"
#include "bubble_generator.h"
#include "time_updateable.h"
#include "templates/callable_back.h"

/**
 * @brief "Glued-on" bubbles in the play field.
 *
 * This class represents "glued-on" bubbles on the rear wall of the play field.
 *
 * Reference object (cannot be copied).
 */
class BubbleField: public TimeUpdateable, public GraphicalNode,
	public Templates::CallableBack<Bubble>
{
	private:

		/**
		 * @brief Class that encapsulates coordinates used in BubbleField
		 */
		class Coordinates
		{
			public:

				/// Plate index
				unsigned plate;
				/// X coordinate
				unsigned x;
				/// Y coordinate
				unsigned y;

			public:

				/// Constructor
				Coordinates(unsigned _plate, unsigned _x, unsigned _y):
					plate(_plate), x(_x), y(_y)
				{
				}

				/**
				 * @brief Implements function less than
				 *
				 * @param[in]  rhs  Right handed side
				 *
				 * @returns Look to the implementation what it returns
				 *
				 * Implemented in order to be able to include Coordinates to std::set.
				 * It is asymetric, transitive and ireflexive.
				 *
				 */
				bool operator < (const Coordinates& rhs) const
				{
					if (plate == rhs.plate) {
						if (x == rhs.x) {
							return y < rhs.y;
						} else {
							return x < rhs.x;
						}
					} else {
						return plate < rhs.plate;
					}
				}

				/**
				 * @brief Comparison operator
				 *
				 * @param[in]  rhs  Right hand side
				 */
				bool operator ==  (const Coordinates& rhs) const
				{
					return plate == rhs.plate && x == rhs.x && y == rhs.y;
				}

				/**
				 * @brief Comparison operator
				 *
				 * @param[in]  rhs  Right hand side
				 */
				bool operator !=  (const Coordinates& rhs) const
				{
					return !(*this == rhs);
				}
		};

		/**
		 * @brief Functor that returns true if given position is outside BubbleField area
		 */
		class isInvalidPosition
		{
			private:

				/// Reference to BubbleField
				const BubbleField& _bubbleField;

			public:

				/// Constructor
				isInvalidPosition(const BubbleField& bubbleField):
					_bubbleField(bubbleField)
				{
				}

				/// Implements functor behaviour
				bool operator() (const Coordinates& coords)
				{
					return !_bubbleField.isValidPosition(coords.plate, coords.x, coords.y);
				}

		};

		/**
		 * @brief Stores Bubbles that are going to be removed because they were not
		 * connected with rear wall
		 */
		class FallingBubblesStorage: public TimeUpdateable, public GraphicalNode
		{
			private:

				/// Typedef to storage type
				typedef std::list<Bubble *> StorageType;

				/// Bubbles in the storage
				StorageType _storage;

			public:

				FallingBubblesStorage();

				~FallingBubblesStorage();

				// Overrides
				virtual void timeUpdate(float dt);

				/**
				 * @brief adds given Bubble to storage
				 *
				 * @param[in]  bubble  Given Bubble
				 */
				void addBubble(Bubble* bubble);

		};

	private:

		/// Number of bubbles in one row
		static const unsigned BUBBLES_IN_ROW;

		/// Types used in for the field content
		typedef Bubble*** BubblePlate;
		typedef std::vector<BubblePlate> FieldContent;

		/// Content of the field
		FieldContent _content;

		/// Object on which the "action performed" method should be called when
		/// a bubble is added to the field
		Templates::CallableBack<BubbleField>* _addBubbleDone;

		/// Container to store bubbles that are exploding at the moment
		std::list<Bubble *> _explodingBubbles;

		/// Stores the number of bubbles that are exploding at the moment
		int _explodingBubblesCount;

		/// Stores falling bubbles because they are not connected to the rear wall
		FallingBubblesStorage _fallingBubbles;

		/// Function that should be called on successfull game end
		void (*_successGameEndFunc)();

	private:

		// Disable copy constructor and assignment operator
		BubbleField(const BubbleField &);
		BubbleField & operator=(const BubbleField &);

		/**
		 * @brief Returns number of Bubbles in one row (or column) in given plate
		 *
		 * @param[in]  index  Plate index
		 */
		unsigned bubblesInPlateRow(unsigned index) const;

		/**
		 * @brief Returns true if given coordinates are valid in the Bubble Field
		 *
		 * @param[in]  plate  Plate index
		 * @param[in]  x      X-axis coordinate
		 * @param[in]  y      Y-axis coordinate
		 */
		bool isValidPosition(unsigned plate, unsigned x, unsigned y) const;

		/**
		 * @brief Returns true if the bubble is in collision with the rear wall
		 *
		 * @param[in]  position  Vector representing position of the bubble
		 */
		static bool isInCollisionWithWall(const SbVec3f& position);

		/**
		 * @brief Returns true if bubble collides in given plate
		 *
		 * @param[in]  plate     Index of the plate to be checked
		 * @param[in]  position  Vector representing position of the bubble
		 */
		bool bubbleCollidesInPlate(unsigned plate, const SbVec3f& position) const;

		/**
		 * @brief Adds bubbles colliding to given bubble to given vector
		 *
		 * @param[in]   plate   Index of the plate to be checked
		 * @param[in]   bubble  Given bubble
		 * @param[out]  collidingBubbles  Vector where the bubbles should be placed
		 */
		void addCollidingBubblesInPlateToVector(unsigned plate, const Bubble* bubble,
				std::vector<const Bubble*>& collidingBubbles) const;

		/**
		 * @brief Solves collision of the given bubble with the wall
		 *
		 * @param[in]  bubble  Given bubble
		 */
		void solveCollisionWithWall(Bubble* bubble);

		/**
		 * @brief Solves collision of the given bubble with the wall
		 *
		 * @param[in]  bubble  Given bubble
		 */
		void solveCollisionWithBubble(Bubble* bubble);

		/**
		 * @brief Remove bubbles of the same type in one cluster
		 *
		 * @param[in]  position  Position where was last bubble added
		 */
		void destroyBubblesInCluster(const Coordinates& position);

		/**
		 * @brief Selects bubbles not connected with the rear wall and removes them
		 */
		void removeBubblesNotConnectedWithWall();

		/**
		 * @brief Searches for nearest suitable position for given bubble
		 *
		 * @param[in]  bubble  Given bubble
		 */
		Coordinates searchForPosition(Bubble* bubble) const;

		/**
		 * @brief Returns vector of bubbles that are in collision with given bubble
		 *
		 * @param[in]  bubble Given bubble
		 *
		 * @returns Returns vector of bubbles with which given bubble is in collision
		 */
		std::vector<const Bubble*> getCollidingBubbles(Bubble* bubble) const;

		/**
		 * @brief Returns list of Coordinates that are neighbours' positions
		 *
		 * @param[in]  coords  Coordinates of given position
		 *
		 */
		std::list<Coordinates> getNeighboursCoords(const Coordinates& coords) const;

		/**
		 * @brief Adds the selected bubble to the selected location
		 *
		 * @param[in] bubble Bubble ot be added
		 * @param[in] plate  Plate coordinate to the field
		 * @param[in] x      X coordinate to the field
		 * @param[in] y      Y coordinate to the field
		 *
		 * Selected location have to be valid and there must not already be a bubble.
		 */
		void addBubble(Bubble* bubble, unsigned plate, unsigned x, unsigned y);

		/**
		 * @brief Recursively adds neighbours towards the rear wall until the wall is hit
		 *
		 * @param[in] bubbleGen  Bubble Generator
		 * @param[in] coords     Coordinates of a bubble for which this function
		 *                       adds neighbours towards wall
		 */
		void addNeighboursTowardsWall(const BubbleGenerator* bubbleGen,
			const Coordinates& coords);

		/**
		 * @brief Returns true if the bubble at the selected coordinates is connected
		 *        to the rear wall, false otherwise
		 *
		 * @param[in]  coords              Current bubble coordinates
		 * @param[in]  processedCoords     Already checked coordinates (to prevent
		 *                                 endless recursion)
		 * @param[out] nearestToWallCoords Coordinates which are nearest to the wall
		 */
		bool isConnectedWithWall(const Coordinates& coords,
				std::vector<Coordinates> processedCoords,
				Coordinates& nearestToWallCoords) const;

		/**
		 * @brief Returns true if the bubble at the selected coordinates is connected
		 *        to the rear wall, false otherwise. Wrapper around
		 *        isConnectedWithWall(const Coordinates&, std::vector<Coordinates>,
		 *        Coordinates&) const
		 *
		 * @param[in]  coords              Current bubble coordinates
		 */
		bool isConnectedWithWall(const Coordinates& coords) const;

		/**
		 * @brief Transforms selected OI coordinate (X or Y) into a local coordinate
		 *
		 * @param[in]  oiCoord    OI X (or Y) coordinate value
		 * @param[in]  plate      Plate index
		 * @param[in]  roundFunc  Rounding function (roundf, floor, ceil, ...)
		 * @return Local X (or Y) coordinate
		 *
		 * X and Y coordinates have always the same value.
		 */
		static unsigned oiXYCoordinateToLocal(float oiCoord, unsigned plate,
				float (*roundFunc)(float) = roundf);

		/**
		 * @brief Transforms z-axis OI coordinate into a local coordinate
		 *
		 * @param[in]  oiCoord    Z coordinate value
		 * @param[in]  roundFunc  Rounding function (roundf, floor, ceil, ...)
		 * @return Local X Z coordinate
		 */
		static unsigned oiZCoordinateToLocal(float oiCoord, float (*roundFunc)(float) = roundf);

		/**
		 * @brief Transforms local coordinates to these in OI
		 *
		 * @param[in]  plate  Plate index
		 * @param[in]  x      Local X coordinate
		 * @param[in]  y      Local Y coordinate
		 * @return Coordinates in OI
		 */
		static SbVec3f localCoordinatesToOI(unsigned plate, unsigned x, unsigned y);

		/**
		 * @brief Transforms local coordinates to these in OI
		 *
		 * @param[in]  local  Local coordinates
		 * @return Coordinates in OI
		 *
		 * Calls localCoordinatesToOI(unsigned plate, unsigned x, unsigned y)
		 *
		 */
		static SbVec3f localCoordinatesToOI(Coordinates local);

	public:

		/**
		 * @brief Default constructor.
		 */
		BubbleField();

		/**
		 * @brief Destructor.
		 */
		~BubbleField();

		/**
		 * @brief Set callback that should be called on successfull game end.
		 *
		 * @param callback Callback to be set.
		 *
		 * This function should be called after the object is created
		 * (if it is not called, the game will continue even if the player
		 * should have won).
		 */
		void setOnSuccessfullGameEndCallback(void (*callback)());

		/**
		 * @brief Generates starting layout
		 *
		 * @param[in] bubbleGen                  Bubble Generator
		 * @param[in] numOfInitiallyFilledPlates Number of initially filled
		 *                                       plates with bubbles
		 * @param[in] probCellContainBubble      Probability that a single cell
		 *                                       will contain a bubble [0.0 - 1.0]
		 *
		 * This method should be called only once and before the game starts.
		 */
		void generateStartingLayout(const BubbleGenerator* bubbleGen,
			unsigned numOfInitiallyFilledPlates, float probCellContainBubble);

		/**
		 * @brief Adds bubble to the field
		 *
		 * @param[in]  bubble         Bubble to be added to the field
		 * @param[in]  addBubbleDone  Object on which the "action performed" method
		 *                            should be called when the bubble is
		 *                            successfully added
		 *
		 * Adds given bubble to the playfield. The given bubble should have
		 * coordinates in bubble field coordinates system for which method
		 * isInCollision() returns true
		 */
		void addBubble(Bubble* bubble, Templates::CallableBack<BubbleField>* addBubbleDone);

		/**
		 * @brief Check for a collision
		 *
		 * @param[in]  position  Position of a bubble which should be checked if
		 * it hasn't collided. Note that the position vector coordinates are
		 * treated as if they were in bubble field coordinates system.
		 *
		 * Returnes if the bubble on given coordinates has collided
		 */
		bool isInCollision(const SbVec3f& position);

		/**
		 * @brief Returns the number of bubbles in one row
		 */
		static unsigned getNumberOfBubblesInRow() { return BUBBLES_IN_ROW; }

		// Override
		virtual void timeUpdate(float dt);

		// Override
		virtual void callableBackNotify(Bubble *subject);
};

#endif
