/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    bubble_generator.h
 * @brief   BubbleGenerator class declaration.
 * @date    2008
 *
 */

#ifndef _BUBBLE_GENERATOR_H_
#define _BUBBLE_GENERATOR_H_

#include <vector>

#include "bubble.h"
#include "camera.h"

/**
 * @brief Bubble generator.
 *
 * This class implements bubble generation according to passed parameters.
 *
 * Reference object (cannot be copied).
 */
class BubbleGenerator
{
	private:

		/// Container to store bubbles.
		typedef std::vector<Bubble *> BubbleContainerType;

		/// Available bubbles to be generated.
		BubbleContainerType _availableBubbles;

		/// Main camera.
		Camera *_camera;

		/// Number of different bubbles that the generateBubble() method can generate.
		const unsigned NUMBER_OF_DIFFERENT_BUBBLES;

	private:

		// Disable copy constructor and assignment operator
		BubbleGenerator(const BubbleGenerator &);
		BubbleGenerator & operator=(const BubbleGenerator &);

		/**
		 * @brief Generates all available bubbles and stores them into _availableBubbles.
		 */
		void generateAvailableBubbles();

		/**
		 * @brief Returns random index to the _availableBubbles container.
		 *
		 * 0 <= getRandomIndex() < NUMBER_OF_DIFFERENT_BUBBLES
		 */
		unsigned getRandomIndex() const;

	public:

		/**
		 * @brief Constructor.
		 *
		 * @param[in] numOfDifferentBubbles Number of different bubbles that
		 *                                  this object can generate.
		 * @param[in] camera                Main camera (needed by some types
		 *                                  of explosions).
		 */
		BubbleGenerator(unsigned numOfDifferentBubbles, Camera *camera);

		/**
		 * @brief Destructor.
		 */
		~BubbleGenerator();

		/**
		 * @brief Generates a single bubble.
		 *
		 * @return Created bubble. If there are no bubbles available, returns
		 *         null pointer.
		 *
		 * Generate bubble is allocated on the heap so it have to be deleted
		 * manually.
		 */
		Bubble *generateBubble() const;
};

#endif
