/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    camera.h
 * @brief   Camera class declaration.
 * @date    2008
 *
 */


#ifndef _CAMERA_H_
#define _CAMERA_H_

#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoRotationXYZ.h>

#include "graphical_node.h"
#include "time_updateable.h"
#include "templates/observable.h"

/**
 * @brief Camera.
 *
 * Reference object (cannot be copied).
 */
class Camera: public TimeUpdateable, public Templates::Observable<Camera>
{
	private:

		/// The root node for the camera composition.
		SoGroup *_rootNode;

		/// Rotation in the X axis.
		SoRotationXYZ *_rotX;

		/// Rotation in the Y axis.
		SoRotationXYZ *_rotY;

		/// X angle (in radians).
		float _angleX;

		/// Y angle (in radians).
		float _angleY;

		/// X angle constrain (in radians).
		float _maxAngleX;

		/// Y angle constrain (in radians).
		float _maxAngleY;

	private:

		// Disable copy constructor and assignment operator
		Camera(const Camera &);
		Camera & operator=(const Camera &);

	public:

		/**
		 * @brief Default constructor.
		 */
		Camera();

		/**
		 * @brief Destructor.
		 */
		~Camera();

	public:

		/**
		 * @brief Returns angle X.
		 */
		float getAngleX() const { return _angleX; }

		/**
		 * @brief Returns angle Y.
		 */
		float getAngleY() const { return _angleY; }

		/**
		 * @brief Returns the root node of the camera object.
		 */
		SoGroup *getRootNode() { return _rootNode; }

		/**
		 * @brief Returns actual camera rotation in the X axis.
		 */
		const SoRotationXYZ *getRotationX() const { return _rotX; }

		/**
		 * @brief Returns actual camera rotation in the Y axis.
		 */
		const SoRotationXYZ *getRotationY() const { return _rotY; }

		// Override
		virtual void timeUpdate(float dt);

		/**
		 * @brief Updates camera view according to the new cursor position difference.
		 *
		 * @param[in] x Difference in the x-axis.
		 * @param[in] y Difference in the y-axis.
		 */
		void updateView(short x, short y);
};

#endif
