/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    explosion.h
 * @brief   Explosion class declaration.
 * @date    2008
 *
 */

#ifndef _EXPLOSION_H_
#define _EXPLOSION_H_

#include "bubble.h"
#include "graphical_node.h"
#include "time_updateable.h"
#include "templates/callable_back.h"

/**
 * @brief Bubble explosion effect.
 *
 * Classes that inherit from this class should only override timeUpdate().
 *
 * Reference object (cannot be copied).
 */
class Explosion: public TimeUpdateable, public GraphicalNode
{
	private:

		/// Explosion position.
		SoTranslation *_position;

	private:

		// Disable copy constructor and assignment operator
		Explosion(const Explosion &);
		Explosion & operator=(const Explosion &);

	protected:

		/// Object on which the "action performed" method should be called when
		/// the explosion is completed.
		Templates::CallableBack<Explosion> *_explodeDone;

		/// Explosion colour.
		const SbColor _color;

		/// Is the explosion active?
		bool _isActive;

	public:

		/**
		 * @brief Constructor.
		 *
		 * @param[in] color Explosion color.
		 */
		Explosion(SbColor color);

		/**
		 * @brief Destructor.
		 */
		virtual ~Explosion();

		/**
		 * @brief Returns a copy of itself.
		 *
		 * Memory is allocated on the heap, so it has to be deleted manually.
		 */
		virtual Explosion *clone() const =0;

		/**
		 * @brief Shows a graphical explosion effect of a bubble.
		 *
		 * @param[in] explodeDone Object on which the "action performed" method
		 *                        should be called when the explosion is completed.
		 *
		 * Effect will be "displayed" on the object root node.
		 */
		void explode(Templates::CallableBack<Explosion> *explodeDone);

		/**
		 * @brief Sets explosion position.
		 *
		 * @param position Explosion position.
		 *
		 * This method should be called before explode() to achieve
		 * proper explosion position.
		 */
		void setPosition(SbVec3f position)
		{
			_position->translation.setValue(position);
		}
};

#endif
