/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    fire_explosion.cc
 * @brief   FireExplosion class definition.
 * @date    2008
 *
 * Implementation is based on http://www.root.cz/clanky/hra-tanky-v-open-inventoru/.
 */

#include <cstdlib>
#include <string>

#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTransparencyType.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <Inventor/nodes/SoCoordinate3.h>

#include "fire_explosion.h"
#include "path_functions.h"

// NOTE: I'm sorry if I left some magical constants in the code, but
// it was based on a foreign code where there were no explanation...

const int FireExplosion::TEXTURE_INDEX_NOT_SET;
const int FireExplosion::TEXTURE_INDEX_MIN;
const int FireExplosion::TEXTURE_INDEX_MAX;

namespace
{
	const char *AVAILABLE_TEXTURE_FILE_NAMES[] =
	{
		"imgs/fire_explosions/explode_0.png",
		"imgs/fire_explosions/explode_1.png",
	};

	const size_t AVAILABLE_TEXTURE_FILE_NAMES_SIZE =
		sizeof(AVAILABLE_TEXTURE_FILE_NAMES) / sizeof(char *);
}

const char *FireExplosion::getRandomTextureFileName()
{
	return AVAILABLE_TEXTURE_FILE_NAMES[static_cast<unsigned>(
		AVAILABLE_TEXTURE_FILE_NAMES_SIZE * (std::rand() / (RAND_MAX + 1.0)))];
}

FireExplosion::FireExplosion(SbColor color, Camera *camera):
	Explosion(color), _actualTime(0.0f), _textureIndex(TEXTURE_INDEX_NOT_SET),
	_billboard(0), _textureCoord(0), _camera(camera), _rotX(0), _rotY(0)
{
	_camera->addObserver(this);

	_rotX = new SoRotationXYZ;
	_rotX->axis = SoRotationXYZ::X;
	_rootNode->addChild(_rotX);

	_rotY = new SoRotationXYZ;
	_rotY->axis = SoRotationXYZ::Y;
	_rootNode->addChild(_rotY);

	_billboard = new SoSeparator;

	SoTransparencyType *transparencyType = new SoTransparencyType;
	transparencyType->value = SoTransparencyType::DELAYED_ADD;
	_billboard->addChild(transparencyType);

	_textureCoord = new SoTextureCoordinate2;
	_billboard->addChild(_textureCoord);
	_textureCoord->point.set1Value(0, SbVec2f(0,        0));
	_textureCoord->point.set1Value(1, SbVec2f(0 + 0.25, 0));
	_textureCoord->point.set1Value(2, SbVec2f(0,        0 + 0.25));
	_textureCoord->point.set1Value(3, SbVec2f(0 + 0.25, 0 + 0.25));

	// Because we're using custom texture colour, we need to set material
	// colour to black...
	SoMaterial *material = new SoMaterial;
	material->diffuseColor.setValue(0.0f, 0.0f, 0.0f);
	material->ambientColor.setValue(0.0f, 0.0f, 0.0f);
	_billboard->addChild(material);

	// ... and disable phong shading
	SoLightModel *lightModel = new SoLightModel;
	lightModel->model = SoLightModel::BASE_COLOR;
	_billboard->addChild(lightModel);

	// Use texture blending with the bubble color
	SoTexture2 *texture = new SoTexture2;
	texture->filename.setValue(PathFunctions::createProperFilePath(
		FireExplosion::getRandomTextureFileName()).c_str());
	texture->model = SoTexture2::BLEND;
		texture->blendColor.setValue(_color);
	_billboard->addChild(texture);

	SbVec3f triangle[4];
	triangle[0].setValue(5, -5, 0);
	triangle[1].setValue(-5, -5, 0);
	triangle[2].setValue(5, 5, 0);
	triangle[3].setValue(-5, 5, 0);

	SoCoordinate3 *stripCoords = new SoCoordinate3;
	SoTriangleStripSet *strip = new SoTriangleStripSet;
	stripCoords->point.setValues(0, 4, &triangle[0]);
	_billboard->addChild(stripCoords);
	strip->numVertices.set1Value(0, 4);
	_billboard->addChild(strip);
}

FireExplosion::~FireExplosion()
{
	_camera->deleteObserver(this);
}

FireExplosion *FireExplosion::clone() const
{
	return new FireExplosion(_color, _camera);
}

void FireExplosion::timeUpdate(float dt)
{
	if (_isActive) {
		_actualTime += dt;

		if (_textureIndex == TEXTURE_INDEX_NOT_SET) {
			// Start explosion
			_rootNode->addChild(_billboard);
			_actualTime = 0.0f;
			_textureIndex = TEXTURE_INDEX_MIN;
		}

		if (_textureIndex > TEXTURE_INDEX_MAX) {
			// End explosion
			_rootNode->removeChild(_billboard);
			_textureIndex = TEXTURE_INDEX_NOT_SET;
			_isActive = false;
			_explodeDone->callableBackNotify(this);
			return;
		}

		if (_textureIndex <= TEXTURE_INDEX_MAX &&
				_actualTime > (_textureIndex * 0.04)) {
			// Process explosion
			int dyt = _textureIndex % 4;
			int dxt = _textureIndex / 4;
			float dx = dyt * 0.25f;
			float dy = 0.75f - (dxt * 0.25f);
			_textureCoord->point.set1Value(0, SbVec2f(dx,         dy));
			_textureCoord->point.set1Value(1, SbVec2f(dx + 0.25f, dy));
			_textureCoord->point.set1Value(2, SbVec2f(dx,         dy + 0.25f));
			_textureCoord->point.set1Value(3, SbVec2f(dx + 0.25f, dy + 0.25f));
			++_textureIndex;
		}
	}
}

void FireExplosion::update(Camera *subject)
{
	// Update billboard rotation according to the camera rotation
	_rotX->angle.setValue(-subject->getRotationX()->angle.getValue());
	_rotY->angle.setValue(-subject->getRotationY()->angle.getValue());
}
