/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    game_menu.cc
 * @brief   GameMenu class definition.
 * @date    2008
 *
 */

#include <cmath>
#include <vector>

#include <qapplication.h>
#include <qpushbutton.h>
#include <qfont.h>
#include <qlabel.h>

#include "game_menu.h"
#include "../factories/factory_environment.h"
#include "../factories/factory_avatar.h"
#include "../factories/factory_play_box_graphics.h"
#include "../factories/factory_play_box_physics.h"

using namespace GUI;

namespace
{
	// NUM_OF_DIFFERENT_BUBBLES 3-10
	// NUM_OF_INITIALLY_FILLED_PLATES 1-10
	// EASY_PROB_CELL_CONTAIN_BUBBLE 1-100 1=0.25-100=1.00
	// EASY_WALL_SPEED 1-100 1=0.01 100=1.00
	const unsigned EASY_NUM_OF_DIFFERENT_BUBBLES = 3;
	const unsigned EASY_NUM_OF_INITIALLY_FILLED_PLATES = 2;
	const unsigned EASY_PROB_CELL_CONTAIN_BUBBLE = 1;
	const unsigned EASY_WALL_SPEED = 1;
	const unsigned MEDIUM_NUM_OF_DIFFERENT_BUBBLES = 5;
	const unsigned MEDIUM_NUM_OF_INITIALLY_FILLED_PLATES = 4;
	const unsigned MEDIUM_PROB_CELL_CONTAIN_BUBBLE = 20;
	const unsigned MEDIUM_WALL_SPEED = 20;
	const unsigned HARD_NUM_OF_DIFFERENT_BUBBLES = 7;
	const unsigned HARD_NUM_OF_INITIALLY_FILLED_PLATES = 6;
	const unsigned HARD_PROB_CELL_CONTAIN_BUBBLE = 40;
	const unsigned HARD_WALL_SPEED = 60;
}

GameMenu::GameMenu(QWidget *parent, const char *name):
	QWidget(parent, name), _grid(0), _environmentCB(0), _avatarCB(0),
	_playBoxGraphicsCB(0), _playBoxPhysicsCB(0), _difficultyCB(0)
{
	// Game menu layout = 15 rows, 8 cols, 15 margin
	_grid = new QGridLayout(this, 15, 8, 15);

	// Environment
	QLabel *label = new QLabel(this);

	label->setText("Environment:");
	_grid->addMultiCellWidget(label, 0, 0, 1, 3);

	_environmentCB = new QComboBox(this);
	std::vector<std::string> registeredEnvironments =
		Factories::FactoryEnvironment::instance().registeredObjects();
	for (std::vector<std::string>::const_iterator iter = registeredEnvironments.begin();
		iter != registeredEnvironments.end(); ++iter) {
		_environmentCB->insertItem(QString(*iter));
	}
	_grid->addMultiCellWidget(_environmentCB, 0, 0, 4, 6);

	// Avatar
	label = new QLabel(this);
	label->setText("Avatar:");
	_grid->addMultiCellWidget(label, 1, 1, 1, 3);

	_avatarCB = new QComboBox(this);
	std::vector<std::string> registeredAvatars =
		Factories::FactoryAvatar::instance().registeredObjects();
	for (std::vector<std::string>::const_iterator iter = registeredAvatars.begin();
		iter != registeredAvatars.end(); ++iter) {
		_avatarCB->insertItem(QString(*iter));
	}
	_grid->addMultiCellWidget(_avatarCB, 1, 1, 4, 6);

	// Play box graphics
	label = new QLabel(this);
	label->setText("Play box graphics:");
	_grid->addMultiCellWidget(label, 2, 2, 1, 3);

	_playBoxGraphicsCB = new QComboBox(this);
	std::vector<std::string> registeredGraphics =
		Factories::FactoryPlayBoxGraphics::instance().registeredObjects();
	for (std::vector<std::string>::const_iterator iter = registeredGraphics.begin();
		iter != registeredGraphics.end(); ++iter) {
		_playBoxGraphicsCB->insertItem(QString(*iter));
	}
	_grid->addMultiCellWidget(_playBoxGraphicsCB, 2, 2, 4, 6);

	// Play box physics
	label = new QLabel(this);
	label->setText("Play box physics:");
	_grid->addMultiCellWidget(label, 3, 3, 1, 3);

	_playBoxPhysicsCB = new QComboBox(this);
	std::vector<std::string> registeredPhysics =
		Factories::FactoryPlayBoxPhysics::instance().registeredObjects();
	for (std::vector<std::string>::const_iterator iter = registeredPhysics.begin();
		iter != registeredPhysics.end(); ++iter) {
		_playBoxPhysicsCB->insertItem(QString(*iter));
	}
	_grid->addMultiCellWidget(_playBoxPhysicsCB, 3, 3, 4, 6);

	// Difficulty
	label = new QLabel(this);
	label->setText("Difficulty:");
	_grid->addMultiCellWidget(label, 4, 4, 1, 3);

	_difficultyCB = new QComboBox(this);
	_difficultyCB->insertItem(QString("Easy"));
	_difficultyCB->insertItem(QString("Medium"));
	_difficultyCB->insertItem(QString("Hard"));
	_difficultyCB->insertItem(QString("Custom"));
	_grid->addMultiCellWidget(_difficultyCB, 4, 4, 4, 6);

	connect(_difficultyCB, SIGNAL(activated (int)),
		this, SLOT(difficultyChanged()));


	// Number of different bubbles
	label = new QLabel(this);
	label->setText("Number of different bubbles:");
	_grid->addMultiCellWidget(label, 5, 5, 1, 3);

	_numberOfDifferentBubblesSL = new QSlider(3, 9, 1, 3, Horizontal, this);
	_numberOfDifferentBubblesSL->setTickmarks(QSlider::Below);
	_numberOfDifferentBubblesSL->setTickInterval(1);
	_grid->addMultiCellWidget(_numberOfDifferentBubblesSL, 5, 5, 4, 6);

	label = new QLabel(this);
	label->setText("3");
	label->setAlignment(Qt::AlignTop || Qt::AlignLeft);
	_grid->addMultiCellWidget(label, 6, 6, 4, 4);
	label = new QLabel(this);
	label->setText("9");
	label->setAlignment(Qt::AlignRight);
	_grid->addMultiCellWidget(label, 6, 6, 6, 6);
	_grid->setRowStretch(6, 0);


	// Number of filled layers
	label = new QLabel(this);
	label->setText("Number of filled layers:");
	_grid->addMultiCellWidget(label, 7, 7, 1, 3);

	_numberOfFilledLayersSL = new QSlider(1, 10, 1, 1, Horizontal, this);
	_numberOfFilledLayersSL->setTickmarks(QSlider::Below);
	_numberOfFilledLayersSL->setTickInterval(1);
	_grid->addMultiCellWidget(_numberOfFilledLayersSL, 7, 7, 4, 6);

	label = new QLabel(this);
	label->setText("1");
	label->setAlignment(Qt::AlignTop || Qt::AlignLeft);
	_grid->addMultiCellWidget(label, 8, 8, 4, 4);
	label = new QLabel(this);
	label->setText("10");
	label->setAlignment(Qt::AlignRight);
	_grid->addMultiCellWidget(label, 8, 8, 6, 6);

	// Bubbles density in one layer
	label = new QLabel(this);
	label->setText("Bubbles density in one layer:");
	_grid->addMultiCellWidget(label, 9, 9, 1, 3);

	_bubblesDensitySL = new QSlider(1, 100, 1, 1, Horizontal, this);
	_bubblesDensitySL->setTickmarks(QSlider::Below);
	_bubblesDensitySL->setTickInterval(10);
	_grid->addMultiCellWidget(_bubblesDensitySL, 9, 9, 4, 6);

	label = new QLabel(this);
	label->setText("low");
	label->setAlignment(Qt::AlignTop || Qt::AlignLeft);
	_grid->addMultiCellWidget(label, 10, 10, 4, 4);
	label = new QLabel(this);
	label->setText("high");
	label->setAlignment(Qt::AlignRight);
	_grid->addMultiCellWidget(label, 10, 10, 6, 6);

	// Rear wall speed
	label = new QLabel(this);
	label->setText("Rear wall speed:");
	_grid->addMultiCellWidget(label, 11, 11, 1, 3);

	_rearWallSpeedSL = new QSlider(1, 100, 1, 1, Horizontal, this);
	_rearWallSpeedSL->setTickmarks(QSlider::Below);
	_rearWallSpeedSL->setTickInterval(10);
	_grid->addMultiCellWidget(_rearWallSpeedSL, 11, 11, 4, 6);

	label = new QLabel(this);
	label->setText("low");
	label->setAlignment(Qt::AlignTop || Qt::AlignLeft);
	_grid->addMultiCellWidget(label, 12, 12, 4, 4);
	label = new QLabel(this);
	label->setText("high");
	label->setAlignment(Qt::AlignRight);
	_grid->addMultiCellWidget(label, 12, 12, 6, 6);

	QPushButton *start = new QPushButton("Start", this, "start");
	start->setFont(QFont("Times", 18, QFont::Bold));

	_grid->setRowStretch(13, 100);

	_grid->addMultiCellWidget(start, 14, 14, 1, 6);

	connect(start, SIGNAL(clicked()), parent, SLOT(startGameNew()));

	// set Medium
	_difficultyCB->setCurrentItem(1);
	difficultyChanged();
}

GameConfiguration GameMenu::getGameConfig()
{
	GameConfiguration gameConfig;

	gameConfig.environmentTypeId = static_cast<const char *>(
			_environmentCB->currentText());
	gameConfig.avatarTypeId = static_cast<const char *>(
			_avatarCB->currentText());
	gameConfig.playBoxGraphicsTypeId = static_cast<const char *>(
			_playBoxGraphicsCB->currentText());
	gameConfig.playBoxPhysicsTypeId = static_cast<const char *>(
			_playBoxPhysicsCB->currentText());

	gameConfig.difficultySettings.numOfDifferentBubbles =
		_numberOfDifferentBubblesSL->value();
	gameConfig.difficultySettings.numOfInitiallyFilledPlates =
		_numberOfFilledLayersSL->value();
	gameConfig.difficultySettings.probCellContainBubble =
		0.25 + ((static_cast<double>(_bubblesDensitySL->value()))/100.0) * 0.75;
	// log((-x*0.9+1))/log(1/10)
	float wallSpeed = log(-static_cast<float>(_rearWallSpeedSL->value()) * 0.9f /
			100.0f + 1.0f) / log(1.0f / 10.0f);
	gameConfig.difficultySettings.wallSpeed = wallSpeed;

	return gameConfig;
}

void GameMenu::difficultyChanged(void)
{
	switch(_difficultyCB->currentItem())
	{
		case 0:
			_numberOfDifferentBubblesSL->setValue(EASY_NUM_OF_DIFFERENT_BUBBLES);
			_numberOfDifferentBubblesSL->setEnabled(false);

			_numberOfFilledLayersSL->setValue(EASY_NUM_OF_INITIALLY_FILLED_PLATES);
			_numberOfFilledLayersSL->setEnabled(false);

			_bubblesDensitySL->setValue(EASY_PROB_CELL_CONTAIN_BUBBLE);
			_bubblesDensitySL->setEnabled(false);

			_rearWallSpeedSL->setValue(EASY_WALL_SPEED);
			_rearWallSpeedSL->setEnabled(false);
			break;

		case 1:
			_numberOfDifferentBubblesSL->setValue(MEDIUM_NUM_OF_DIFFERENT_BUBBLES);
			_numberOfDifferentBubblesSL->setEnabled(false);

			_numberOfFilledLayersSL->setValue(MEDIUM_NUM_OF_INITIALLY_FILLED_PLATES);
			_numberOfFilledLayersSL->setEnabled(false);

			_bubblesDensitySL->setValue(MEDIUM_PROB_CELL_CONTAIN_BUBBLE);
			_bubblesDensitySL->setEnabled(false);

			_rearWallSpeedSL->setValue(MEDIUM_WALL_SPEED);
			_rearWallSpeedSL->setEnabled(false);
			break;

		case 2:
			_numberOfDifferentBubblesSL->setValue(HARD_NUM_OF_DIFFERENT_BUBBLES);
			_numberOfDifferentBubblesSL->setEnabled(false);

			_numberOfFilledLayersSL->setValue(HARD_NUM_OF_INITIALLY_FILLED_PLATES);
			_numberOfFilledLayersSL->setEnabled(false);

			_bubblesDensitySL->setValue(HARD_PROB_CELL_CONTAIN_BUBBLE);
			_bubblesDensitySL->setEnabled(false);

			_rearWallSpeedSL->setValue(HARD_WALL_SPEED);
			_rearWallSpeedSL->setEnabled(false);
			break;

		case 3:
		default:
			_numberOfDifferentBubblesSL->setEnabled(true);
			_numberOfFilledLayersSL->setEnabled(true);
			_bubblesDensitySL->setEnabled(true);
			_rearWallSpeedSL->setEnabled(true);
			break;
	}
}
