/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    main_window.cc
 * @brief   MainWindow class definition. Top level window.
 * @date    2008
 *
 */

#include <qpushbutton.h>
#include <qlabel.h>
#include <Inventor/Qt/SoQt.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qfont.h>
#include <qcursor.h>

#include "main_window.h"
#include "main_menu.h"
#include "render_window.h"
#include "../path_functions.h"

using namespace GUI;

MainWindow::MainWindow(QWidget *parent, const char *name):
	QWidget(parent, name), _grid(0), _mainWidget(0)
{
	// SoQt init (only once)
	SoQt::init(this);

	// Creates main menu
	GUI::MainMenu *MainMenu = new GUI::MainMenu(this, "MainMenu");

	// Creates implicit render window (and start game)
	GameConfiguration gameConfig;
	gameConfig.environmentTypeId = "Space";
	gameConfig.avatarTypeId = "Snowman";
	gameConfig.playBoxGraphicsTypeId = "Basic";
	gameConfig.playBoxPhysicsTypeId = "Basic";
	RenderWindow *render = new RenderWindow(gameConfig, this, "render");

	// Application layout: 2 rows, 1 column
	_grid = new QGridLayout(this, 2, 1, 10);

	// Set 1. row = menu
	_grid->setRowStretch (1, 100);
	_grid->setRowSpacing(0,20);
	_grid->addWidget(MainMenu, 0, 0);

	// Set 2. row = game
	changeMainWidget(render);
}

void MainWindow::changeMainWidget(QWidget *w)
{
	// Release widget and set new widget
	if (_mainWidget) {
		_mainWidget->hide();
		delete _mainWidget;
		_mainWidget = 0;
	}
	_mainWidget = w;
	_grid->addWidget(w, 1, 0);
	w->show();
}

void MainWindow::showHelpRules()
{
	QLabel *label = new QLabel(this);
	label->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	label->setText("Game Rules:\n      "
		"The player is required to remove all bubbles from the playing field "
		"by shooting them with a bubble of the same colour. If three or more "
		"bubbles of the same colour are placed together, they disappear. It may be "
		"useful to bounce bubbles of walls when grouping bubbles together.\n        "
		"The player is limited by time - the rear wall is moving towards "
		"the player and if the player does not remove all bubbles from the playing "
		"field by the time the wall reaches them, the game is over.\n        "
		"The difficulty of the game is determined by settings in the Game->New "
		"menu. There are three predefined difficulties (easy/medium/hard). In "
		"custom option, you can set your own difficulty.");
	label->setAlignment(AlignTop | WordBreak);
	label->setFont(QFont("SansSerif", 10));
	changeMainWidget(label);
}

void MainWindow::showHelpAbout()
{
	QLabel *label = new QLabel(this);
	label->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	label->setText("BubbleShoooooter:\n      "
		"This program was created as a Computer Graphics project at Faculty of "
		"Information Technology, Brno University of Technology, Brno, Czech "
		"Republic. The task was to create a simple computer game using the "
		"OpenInventor library. Our solution is a 3D variation of a popular "
		" GNU/Linux game called Frozen-Bubble (http://www.frozen-bubble.org)."
		"\n\n"
		"Authors:\n        "
		"Ondrej Lengal (xlenga00@stud.fit.vutbr.cz)\n        "
		"Libor Polcak (xpolca03@stud.fit.vutbr.cz)\n        "
		"Boris Prochazka (xproch63@stud.fit.vutbr.cz)\n        "
		"Petr Zemek (xzemek02@stud.fit.vutbr.cz, team leader)\n\n"
		"License:\n        "
		"GNU GPLv2");
	label->setAlignment(AlignTop | WordBreak);
	label->setFont(QFont("SansSerif", 10));
	changeMainWidget(label);
}

void MainWindow::showGameNew()
{
	_gameMenu = new GUI::GameMenu(this, "GameMenu");
	changeMainWidget(_gameMenu);
}

void MainWindow::startGameNew()
{
	GameConfiguration gameConfig = _gameMenu->getGameConfig();
	RenderWindow *render = new RenderWindow(gameConfig, this, "render");
	changeMainWidget(render);

	QPoint centre(width()/2,height()/2);
	QPoint mouseCentre(mapToGlobal(centre));
	QCursor::setPos(mouseCentre);
}

void MainWindow::gameWin()
{
	QLabel *label = new QLabel(this);
	QImage qImage(PathFunctions::createProperFilePath(
		"imgs/win.jpg").c_str());
	qImage = qImage.smoothScale(this->width()-25,this->height());
	label->setPixmap(qImage);
	label->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	changeMainWidget(label);
}

void MainWindow::gameLose()
{
	QLabel *label = new QLabel(this);
	QImage qImage(PathFunctions::createProperFilePath(
		"imgs/gameover.jpg").c_str());
	qImage = qImage.smoothScale(this->width()-25,this->height());
	label->setPixmap(qImage);
	label->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	changeMainWidget(label);
}
