/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    render_window.cc
 * @brief   RenderWindow class definition.
 * @date    2008
 *
 */

#include <cassert>

#include <Inventor/events/SoLocation2Event.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <qcursor.h>

#include "render_window.h"
#include "main_window.h"
#include "../scene_creator.h"
#include "../create_scene_parameters.h"
#include "../factories/factory_environment.h"
#include "../factories/factory_avatar.h"
#include "../factories/factory_play_box_graphics.h"
#include "../factories/factory_play_box_physics.h"

using namespace GUI;

SbTime RenderWindow::lastTime = SbTime::zero();
Scene *RenderWindow::scene = 0;
Camera *RenderWindow::camera = 0;
SoEventCallback *RenderWindow::eventCallback = 0;
MainWindow *RenderWindow::mainWindow = 0;
bool RenderWindow::stopped = false;
short RenderWindow::oldx = -1;
short RenderWindow::oldy = -1;
bool RenderWindow::failedGameEnd = false;
bool RenderWindow::successGameEnd = false;
bool RenderWindow::haveFocus = true;

namespace
{
	// Name of the game
	const char *GAME_NAME = "Bubble shoooooter";
}

#include <iostream>

RenderWindow::RenderWindow(GameConfiguration gameConfig, QWidget *parent,
	const char *name): QWidget(parent, name), _sceneRoot(0),
	_sceneSensor(0), _renderArea(0)
{
	// Parent has to be a MainWindow instance
	assert(dynamic_cast<MainWindow *>(parent) != 0);
	RenderWindow::mainWindow = dynamic_cast<MainWindow *>(parent);

	// User neither won nor failed the game yet
	RenderWindow::failedGameEnd = false;
	RenderWindow::successGameEnd = false;

	// Create scene root
	_sceneRoot = new SoSeparator;
	_sceneRoot->ref();

	// Create a camera
	camera = new Camera();

	// Create a scene
	CreateSceneParameters sceneParams;
	sceneParams.sceneRoot = _sceneRoot;
	sceneParams.camera = camera;
	sceneParams.gameConfig = gameConfig;
	sceneParams.successGameEndFunc = successGameEndCallback;
	sceneParams.failedGameEndFunc = failedGameEndCallback;
	RenderWindow::scene = SceneCreator::createScene(sceneParams);

	// Create event callbacks (mouse motion + mouse right click)
	RenderWindow::eventCallback = new SoEventCallback;
	RenderWindow::eventCallback->addEventCallback(SoLocation2Event::getClassTypeId(),
		RenderWindow::mouseEventCallback, 0);
	RenderWindow::eventCallback->addEventCallback(SoMouseButtonEvent::getClassTypeId(),
		RenderWindow::mouseButtonEventCallback, 0);
	_sceneRoot->insertChild(RenderWindow::eventCallback, 0);

	// Create sensor aka alarm when idle
	_sceneSensor = new SoOneShotSensor(RenderWindow::sceneSensorCallback, 0);
	_sceneSensor->schedule();

	// Create _renderArea to show our OI scene in Qt
	_renderArea = new SoQtRenderArea(this);
	_renderArea->setAlphaChannel(true);
	_renderArea->setAccumulationBuffer(true);
	_renderArea->setSceneGraph(_sceneRoot);
	_renderArea->setTransparencyType(SoGLRenderAction::DELAYED_BLEND);
	_renderArea->setAntialiasing(true, 2);
	_renderArea->setTitle(GAME_NAME);

	// You must enable keyboard focus for a widget if it processes keyboard events
	setFocusPolicy(QWidget::StrongFocus);

	// When the mouse is out of the window, call mouseMoveEvent
	setMouseTracking(true);

	// Center mouse
	QPoint centre(width()/2,height()/2);
	QPoint mouseCentre(mapToGlobal(centre));
	QCursor::setPos(mouseCentre);

	// Set focus
	setFocus();
	RenderWindow::haveFocus = true;

	_renderArea->show();

	// Reset mouse position
	RenderWindow::oldx = -1;
	RenderWindow::oldy = -1;
}

RenderWindow::~RenderWindow()
{
	delete _renderArea;
	_renderArea = 0;

	if (_sceneSensor->isScheduled()) {
		_sceneSensor->unschedule();
	}
	delete _sceneSensor;
	_sceneSensor = 0;

	delete RenderWindow::scene;
	RenderWindow::scene = 0;

	delete camera;
	camera = 0;

	RenderWindow::failedGameEnd = false;
	RenderWindow::successGameEnd = false;
	RenderWindow::mainWindow = 0;

	_sceneRoot->unref();
}

void RenderWindow::successGameEndCallback()
{
	successGameEnd = true;
}

void RenderWindow::failedGameEndCallback()
{
	failedGameEnd = true;
}

void RenderWindow::sceneSensorCallback(void *, SoSensor *sensor)
{
	if (successGameEnd) {
		mainWindow->gameWin();
	} else if (failedGameEnd) {
		mainWindow->gameLose();
	} else {
		SbTime currentTime = SbTime::getTimeOfDay();
		SbTime dTime;

		// First round (we can not determine dt)
		if (RenderWindow::lastTime == SbTime::zero()) {
			RenderWindow::lastTime = currentTime;
			sensor->schedule();
			return;
		}

		// Determine dTime and save currentTime
		dTime = currentTime - RenderWindow::lastTime;
		RenderWindow::lastTime = currentTime;

		// Call scene update
		if (!RenderWindow::stopped) {
			RenderWindow::scene->timeUpdate(dTime.getValue());
		}
		sensor->schedule();
	}
}

void RenderWindow::mouseEventCallback(void *, SoEventCallback *node)
{
	short x, y;
	const SoEvent *event = node->getEvent();

	SbVec2s vec = event->getPosition();
	vec.getValue(x,y);

	// First round (we can not determine dt)
	if (RenderWindow::oldx < 0 || RenderWindow::oldy < 0) {
		 RenderWindow::oldx = x;
		 RenderWindow::oldy = y;
		 return;
	}

	RenderWindow::camera->updateView(-(oldx-x), -(oldy-y));

	RenderWindow::oldx = x;
	RenderWindow::oldy = y;

	node->setHandled();
}

void RenderWindow::mouseButtonEventCallback(void *, SoEventCallback *node)
{
	const SoEvent *event = node->getEvent();

	// Left mouse button
	if (SoMouseButtonEvent::isButtonPressEvent(event, SoMouseButtonEvent::BUTTON1)) {
		if (RenderWindow::haveFocus) {
			// Fire only if we have focus
			RenderWindow::scene->fire();
		}
	}

	node->setHandled();
}

void RenderWindow::keyPressEvent(QKeyEvent *e)
{
	if (e->key() == Qt::Key_Escape && hasFocus()) {
		clearFocus();
		RenderWindow::haveFocus = false;
	}
}

void RenderWindow::mousePressEvent(QMouseEvent *)
{
	if (!hasFocus()) {
		setFocus();
		RenderWindow::haveFocus = true;
	}
}

void RenderWindow::mouseMoveEvent(QMouseEvent *)
{
	if (!RenderWindow::stopped) {
		QPoint centre(width()/2,height()/2);
		QPoint mouseCentre(mapToGlobal(centre));

		QCursor::setPos(mouseCentre);
		RenderWindow::oldx = -1;
		RenderWindow::oldy = -1;
	}
}

void RenderWindow::focusInEvent(QFocusEvent *)
{
	setCursor(QCursor(Qt::BlankCursor));

	RenderWindow::eventCallback->addEventCallback(SoLocation2Event::getClassTypeId(),
		RenderWindow::mouseEventCallback, 0);
	RenderWindow::eventCallback->addEventCallback(SoMouseButtonEvent::getClassTypeId(),
		RenderWindow::mouseButtonEventCallback, 0);

	RenderWindow::stopped = false;
}

void RenderWindow::focusOutEvent(QFocusEvent *)
{
	setCursor(QCursor(Qt::ArrowCursor));

	RenderWindow::eventCallback->removeEventCallback(SoLocation2Event::getClassTypeId(),
		RenderWindow::mouseEventCallback, 0);
	RenderWindow::eventCallback->removeEventCallback(SoMouseButtonEvent::getClassTypeId(),
		RenderWindow::mouseButtonEventCallback, 0);

	RenderWindow::stopped = true;
}
