/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    render_window.h
 * @brief   RenderWindow class declaration.
 * @date    2008
 *
 */

#ifndef _GUI__RENDER_WINDOW_H_
#define _GUI__RENDER_WINDOW_H_

#include <qwidget.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/sensors/SoOneShotSensor.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/Qt/SoQtRenderArea.h>
#include <Inventor/nodes/SoEventCallback.h>

#include "../game_configuration.h"
#include "../camera.h"
#include "../scene.h"

/**
 * @brief Graphical user interface.
 */
namespace GUI
{
	class RenderWindow;
	class MainWindow;
}

/**
 * @brief Render window class. Makes interconnection between Qt and OI.
 *
 * This class acts like connection layer between Qt application and scene, which
 * is created in Open Inventor.
 */
class GUI::RenderWindow: public QWidget
{
	private:

		/// The root of the generated scene.
		SoSeparator *_sceneRoot;

		/// Scene sensor aka scene alarm. Calls sceneSensorCallback.
		SoOneShotSensor *_sceneSensor;

		/// Qt render area.
		SoQtRenderArea *_renderArea;

		/// Widget parent (it has to be a MainWindow instance).
		static MainWindow *mainWindow;

		/// Call back functions (mouse move, mouse click, keyboard click).
		static SoEventCallback *eventCallback;

		/// True if game is stopped.
		static bool stopped;

		/// Camera.
		static Camera *camera;

		/// Scene.
		static Scene *scene;

		/// The last time, when sceneSensorCallback was triggered.
		static SbTime lastTime;

		/// Old x coordinate.
		static short oldx;

		/// Old y coordinate.
		static short oldy;

		/// If true, user failed the game.
		static bool failedGameEnd;

		/// If true, user won the game.
		static bool successGameEnd;

		/// Do we have focus?
		static bool haveFocus;

	public:

		/**
		 * @brief Default constructor.
		 *
		 * @param[in]  gameConfig Game configuration values.
		 * @param[in]  parent     QWidget parent.
		 * @param[in]  name       QWidget name.
		 */
		RenderWindow(GameConfiguration gameConfig, QWidget *parent = 0,
			const char *name = 0);

		/**
		 * @brief Destructor.
		 */
		~RenderWindow();

		/**
		 * @brief Function that should be called on successfull game end.
		 */
		static void successGameEndCallback();

		/**
		 * @brief Function that should be called on failed game end.
		 */
		static void failedGameEndCallback();

		/**
		 * @brief Function which is called by sensor from Open Inventor.
		 *
		 * @param[in]  userdata Not used.
		 * @param[in]  sensor   Sensor pointer. Used to schedule itself.
		 */
		static void sceneSensorCallback(void *userdata, SoSensor *sensor);

		/**
		 * @brief Function which is called by mouse motion.
		 *
		 * @param[in]  userdata Not used.
		 * @param[in]  node     Event node pointer. Used to determine mouse position.
		 */
		static void mouseEventCallback(void *userdata, SoEventCallback *node);

		/**
		 * @brief Function which is called by mouse click.
		 *
		 * @param[in]  userdata Not used.
		 * @param[in]  node     Event node pointer. Used to determine mouse click.
		 */
		static void mouseButtonEventCallback(void *userdata, SoEventCallback *node);

		/**
		 * @brief Qt key press function.
		 *
		 * @param[in]  e Event.
		 */
		void keyPressEvent (QKeyEvent *e);

		/**
		 * @brief Qt mouse move function.
		 *
		 * @param[in]  e Event.
		 */
		void mouseMoveEvent(QMouseEvent *e);

		/**
		 * @brief Qt mouse press function.
		 *
		 * @param[in]  e Event.
		 */
		void mousePressEvent(QMouseEvent *e);

		/**
		 * @brief Qt in focus function.
		 *
		 * @param[in]  e Event.
		 */
		void focusInEvent(QFocusEvent *e);

		/**
		 * @brief Qt out focus function.
		 *
		 * @param[in]  e Event.
		 */
		void focusOutEvent(QFocusEvent *e);
};

#endif
