/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    model_avatar.cc
 * @brief   Implementation of the model avatar
 * @date    2008
 *
 * Implementation for the model avatar.
 */

#include <cassert>
#include <string>

#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCylinder.h>

#include "model_avatar.h"
#include "factories/factory_avatar.h"
#include "play_field.h"
#include "texture_generator.h"

const float ModelAvatar::INITIAL_HAMMER_ROT = M_PI / 2.0f;

ModelAvatar::ModelAvatar():
	_rotX(0), _rotY(0), _hammerTimer(0), _timerTimeOut(0), _state(STOPPED),
	_fireDone(0)
{
	// initilize the root node
	_rootNode = new SoSeparator();

	// initialize rotations
	_rotX = new SoRotationXYZ;
	_rotX->axis = SoRotationXYZ::X;
	_rootNode->addChild(_rotX);

	_rotY = new SoRotationXYZ;
	_rotY->axis = SoRotationXYZ::Y;
	_rootNode->addChild(_rotY);

	// create the hammer
	SoSeparator* hammerSep = new SoSeparator();

	// move the hammer in the Z-axix
	SoTranslation* hammerTrans = new SoTranslation();
	hammerTrans->translation.setValue(
		SbVec3f(0.0f, 0.0f, -2.5f));
	hammerSep->addChild(hammerTrans);

	// initial rotation of the hammer
	SoRotationXYZ* hammerInitRotation = new SoRotationXYZ();
	hammerInitRotation->axis.setValue(SoRotationXYZ::X);
	hammerInitRotation->angle.setValue(-INITIAL_HAMMER_ROT);
	hammerSep->addChild(hammerInitRotation);

	_hammerTimer = new SoElapsedTime();
	_hammerTimer->speed.setValue(8.0f);
	_hammerTimer->on.setValue(false);

	// create the timer output
	_timerTimeOut = new SoSFFloat();
	_timerTimeOut->connectFrom(&(_hammerTimer->timeOut));

	// rotate the hammer
	SoRotationXYZ* _hammerRotation = new SoRotationXYZ();
	_hammerRotation->axis.setValue(SoRotationXYZ::X);
	_hammerRotation->angle.connectFrom(&(_hammerTimer->timeOut));
	hammerSep->addChild(_hammerRotation);

	// move the hammer in the Y-axix
	hammerTrans = new SoTranslation();
	hammerTrans->translation.setValue(
		SbVec3f(0.0f, PlayField::getSizeHalfSize(), 0.0f));
	hammerSep->addChild(hammerTrans);

	hammerSep->addChild(Avatar::getBubbleHammer());

	_rootNode->addChild(hammerSep);

	// move the snowman
	SoTranslation* snowmanTrans = new SoTranslation();
	snowmanTrans->translation.setValue(SbVec3f(-5.0f, 0.0f, 0.0f));
	_rootNode->addChild(snowmanTrans);

	// invert the coordinate system for the snowman
	SoScale* snowmanScale = new SoScale();
	snowmanScale->scaleFactor.setValue(SbVec3f(1.0f, 1.0f, -1.0f));
	_rootNode->addChild(snowmanScale);

	// constant rotation of the snowman
	SoRotationXYZ* rotSnowman = new SoRotationXYZ();
	rotSnowman->axis.setValue(SoRotationXYZ::X);
	rotSnowman->angle.setValue(-M_PI / 2.0f);
	_rootNode->addChild(rotSnowman);

	rotSnowman = new SoRotationXYZ();
	rotSnowman->axis.setValue(SoRotationXYZ::Z);
	rotSnowman->angle.setValue(INITIAL_HAMMER_ROT);
	_rootNode->addChild(rotSnowman);

	// rotate the snowman with the hammer
	rotSnowman = new SoRotationXYZ();
	rotSnowman->axis.setValue(SoRotationXYZ::Z);
	rotSnowman->angle.connectFrom(&(_hammerTimer->timeOut));
	_rootNode->addChild(rotSnowman);

	// create the avatar (WRL model)
	_rootNode->addChild(TextureGenerator::generateModel("Tux"));
}

ModelAvatar::~ModelAvatar()
{
}

void ModelAvatar::timeUpdate(float /*dt*/)
{
	switch (_state)	{
		case STOPPED: {
				break;
			}

		case PREFIRE: {
				if (fmod(_timerTimeOut->getValue(), 2.0f * M_PI) > INITIAL_HAMMER_ROT) {
					// if hammer touches the bubble
					_state = POSTFIRE;
					_fireDone->callableBackNotify(this);
				}

				break;
			}

		case POSTFIRE: {
				if (fmod(_timerTimeOut->getValue(), 2.0f * M_PI) < INITIAL_HAMMER_ROT) {
					// if hammer touches the bubble
					_state = STOPPED;
					_hammerTimer->on.setValue(false);
				}

				break;
			}

		default:
			assert(false); // Program should never get there (invalid state)
	}
}

void ModelAvatar::update(Camera * subject)
{
	// Update avatar rotation according to the camera rotation
	_rotX->angle.setValue(-subject->getRotationX()->angle.getValue());
	_rotY->angle.setValue(-subject->getRotationY()->angle.getValue());
}

void ModelAvatar::fire(Templates::CallableBack<Avatar>* fireDone)
{
	// start the firing sequence
	// set start values
	_state = PREFIRE;
	_fireDone = fireDone;

	// initialize and unpause the hammer timer
	_hammerTimer->on.setValue(true);
}

// Functions and constants needed by Factories::FactoryAvatar
namespace
{
	// Avatar identifier for the avatars factory
	const std::string MODEL_AVATAR_ID = "Tux";

	// Function that returns an instance of the Avatar class
	Avatar *getModelAvatarInstance()
	{
		return new ModelAvatar::ModelAvatar();
	}

	// Register the selected avatar at the avatars factory
	const bool MODEL_AVATAR_REGISTERED = Factories::FactoryAvatar::instance().
		registerObject(MODEL_AVATAR_ID, getModelAvatarInstance);
}

