/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    play_field.cc
 * @brief   PlayField class definition.
 * @date    2008
 *
 */

#include <algorithm>
#include <cassert>

#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/nodes/SoScale.h>

#include "play_field.h"
#include "play_box_graphics.h"
#include "play_box_physics.h"
#include "bubble.h"
#include "bubble_field.h"
#include "bubble_generator.h"

const float PlayField::PLAY_BOX_SIDE_HALF_SIZE = 6.0f;
const float PlayField::PLAY_BOX_DEPTH = 50.0f;
const float PlayField::PLAY_BOX_NEAR_WALL_BOUND = 2.0f * Bubble::getRadius();
const float PlayField::DEFAULT_WALL_OFFSET = PlayField::PLAY_BOX_DEPTH;

PlayField::PlayField(const SceneConfiguration &sceneConfig):
	_graphics(0), _physics(0), _avatar(0), _camera(0), _bubbleField(0),
	_flyingBubble(0), _bubbleIsReady(false), _bubbleGenerator(0),
	_wallOffset(0.0f), _wallTranslation(0), _wallSpeed(0.0f),
	_failedGameEndFunc(0)
{
	// Initialize the root node
	_rootNode = new SoSeparator();

	// TODO: Should we use objects that count their references?
	assert(sceneConfig.playBoxGraphics != 0);
	assert(sceneConfig.playBoxPhysics  != 0);
	assert(sceneConfig.avatar          != 0);

	// Populate private data members
	_graphics = sceneConfig.playBoxGraphics;
	_physics = sceneConfig.playBoxPhysics;
	_avatar = sceneConfig.avatar;
	_camera = sceneConfig.camera;
	_wallSpeed = sceneConfig.gameDifficulty.wallSpeed;
	_failedGameEndFunc = sceneConfig.failedGameEndFunc;

	// Insert the PlayBoxGraphics
	_rootNode->addChild(_graphics->getRootNode());

	// Move avatar to the proper position (right above the floor - the avatar
	// is by default placed in the middle of the rear side).
	SoSeparator *movedAvatarNode = new SoSeparator();
	SoTranslation *avatarMovement = new SoTranslation();
	avatarMovement->translation.setValue(
		SbVec3f(0.0f, -PLAY_BOX_SIDE_HALF_SIZE, 0.0f));
	movedAvatarNode->addChild(avatarMovement);
	movedAvatarNode->addChild(_avatar->getRootNode());
	_rootNode->addChild(movedAvatarNode);

	_wallOffset = PlayField::DEFAULT_WALL_OFFSET;
	_graphics->moveWall(_wallOffset);

	// Create the bubble generator
	_bubbleGenerator = new BubbleGenerator(
		sceneConfig.gameDifficulty.numOfDifferentBubbles,
		sceneConfig.camera);

	// Create the bubble field
	_bubbleField = new BubbleField();
	_bubbleField->setOnSuccessfullGameEndCallback(
		sceneConfig.successGameEndFunc);
	_bubbleField->generateStartingLayout(_bubbleGenerator,
		sceneConfig.gameDifficulty.numOfInitiallyFilledPlates,
		sceneConfig.gameDifficulty.probCellContainBubble);

	// Get the bubble field generator
	SoSeparator *bubbleFieldSep = new SoSeparator();

	// Create the translation object for the bubble field
	_wallTranslation = new SoTranslation();
	_wallTranslation->translation.setValue(0.0f, 0.0f, _wallOffset);
	bubbleFieldSep->addChild(_wallTranslation);

	SoScale *bubbleFieldScale = new SoScale();
	bubbleFieldScale->scaleFactor.setValue(SbVec3f(1.0f, 1.0f, -1.0f));
	bubbleFieldSep->addChild(bubbleFieldScale);

	bubbleFieldSep->addChild(_bubbleField->getRootNode());

	_rootNode->addChild(bubbleFieldSep);

	createNewBubble();
}

PlayField::~PlayField()
{
	delete _bubbleField;
	_bubbleField = 0;
	delete _bubbleGenerator;
	_bubbleGenerator = 0;
	delete _flyingBubble;
	_flyingBubble = 0;
}

void PlayField::createNewBubble()
{
	assert(_flyingBubble == 0);

	if (_bubbleIsReady) {
		// This means that the user failed and the game ends unsuccessfully
		// for him
		if (_failedGameEndFunc != 0) {
			_failedGameEndFunc();
		}
		return;
	}

	_flyingBubble = _bubbleGenerator->generateBubble();
	_rootNode->addChild(_flyingBubble->getRootNode());
	_flyingBubble->setMotionVector(SbVec3f(0.0f, 0.0f, 0.0f));
	_flyingBubble->enableTransparency();
	_bubbleIsReady = true;
}

SbVec3f PlayField::coordsInBubbleField(const SbVec3f& pfCoords)
{
	return SbVec3f(pfCoords[0], pfCoords[1], _wallOffset - pfCoords[2]);
}

void PlayField::timeUpdate(float dt)
{
	// Compute new back wall offset and move the wall
	_wallOffset -= _wallSpeed * dt;
	if (_wallOffset < PlayField::PLAY_BOX_NEAR_WALL_BOUND) {
		_wallOffset = PlayField::PLAY_BOX_NEAR_WALL_BOUND;
	}
	_wallTranslation->translation.setValue(0.0f, 0.0f, _wallOffset);
	_graphics->moveWall(_wallOffset);

	// Forward time update event
	_graphics->timeUpdate(dt);
	_avatar->timeUpdate(dt);
	_physics->timeUpdate(dt);
	_bubbleField->timeUpdate(dt);

	if (_flyingBubble != 0) {
		const float BUBBLE_RADIUS = _flyingBubble->getRadius();
		float distance = dt * _flyingBubble->getSpeed();
		// Move the bubble few steps to ensure to detect possible collisions
		while (distance > 0.0f) {
			float move = std::min(BUBBLE_RADIUS, distance);
			_physics->moveBubble(_flyingBubble, move);
			// Detect collision in the bubble field
			const SbVec3f posOfBubbleInBubbleField = coordsInBubbleField(
					_flyingBubble->getPosition());
			if (_bubbleField->isInCollision(posOfBubbleInBubbleField)) {
				// Set position of the bubble to the position in bubble field
				// and add it there
				_flyingBubble->setPosition(posOfBubbleInBubbleField);
				// In case the callableBackNotify() function is called
				// before the _bubbleField->addBubble() returns, pass
				// a temporary bubble instead it and set _flyingBubble to 0
				// to ensure proper behaviour later
				Bubble *tmpBubble = _flyingBubble;
				_flyingBubble = 0;
				_bubbleField->addBubble(tmpBubble, this);
				// Forget about last bubble and create new one
				_rootNode->removeChild(tmpBubble->getRootNode());
				break; // Escape from the while loop
			}
			// Prepare for the next step
			distance -= BUBBLE_RADIUS;
		}
	}
}

void PlayField::callableBackNotify(BubbleField */*subject*/)
{
	createNewBubble();
}

void PlayField::callableBackNotify(Avatar */*subject*/)
{
	// When the avatar firing sequence is done, launch the bubble
	launchBubble();
}

void PlayField::fire()
{
	if (_bubbleIsReady) {
		_bubbleIsReady = false;
		_avatar->fire(this);
	}
}

void PlayField::launchBubble()
{
	_flyingBubble->disableTransparency();
	_flyingBubble->setMotionVector(SbVec3f(
				-_camera->getAngleY(), _camera->getAngleX(), 1.0f));
}
