/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    play_field.h
 * @brief   PlayField class declaration.
 * @date    2008
 *
 */

#ifndef _PLAY_FIELD_H_
#define _PLAY_FIELD_H_

#include <Inventor/fields/SoSFTime.h>
#include <Inventor/nodes/SoTranslation.h>

#include "graphical_node.h"
#include "time_updateable.h"
#include "scene_configuration.h"
#include "templates/callable_back.h"

class Avatar;
class PlayBoxGraphics;
class PlayBoxPhysics;
class Bubble;
class BubbleGenerator;
class BubbleField;

/**
 * @brief Play field.
 *
 * Reference object (cannot be copied).
 */
class PlayField: public TimeUpdateable, public GraphicalNode,
	public Templates::CallableBack<Avatar>,
	public Templates::CallableBack<BubbleField>
{
	private:

		/// Half size of a single side of the play box
		static const float PLAY_BOX_SIDE_HALF_SIZE;

		/// Depth of the play box
		static const float PLAY_BOX_DEPTH;

		/// The near bound of the back wall
		static const float PLAY_BOX_NEAR_WALL_BOUND;

		/// Default rear wall offset
		static const float DEFAULT_WALL_OFFSET;

		/// Strategy object that encapsulates an algorithm rendering the play field
		PlayBoxGraphics *_graphics;

		/// Strategy object that encapsulates an algorithm in charge of physics
		PlayBoxPhysics *_physics;

		/// Graphical representation of the player (the avatar)
		Avatar *_avatar;

		/// Camera
		Camera *_camera;

		/// "Glued-on" bubbles on the rear wall
		BubbleField *_bubbleField;

		/// Bubble flying in the field
		Bubble *_flyingBubble;

		/// Interface is waiting for a command to fire the bubble
		bool _bubbleIsReady;

		/// Bubble generator
		BubbleGenerator *_bubbleGenerator;

		/// Offset of the rear wall
		float _wallOffset;

		/// Translation of the rear wall
		SoTranslation* _wallTranslation;

		/// Rear wall speed
		float _wallSpeed;

		/// Function that should be called on failed game end
		void (*_failedGameEndFunc)();

	private:

		// Disable copy constructor and assignment operator
		PlayField(const PlayField &);
		PlayField & operator=(const PlayField &);

		/**
		 * @brief Adds non moving bubble which can be fired
		 */
		void createNewBubble();

		/**
		 * @brief Translates coordinates from play field system to bubble field system
		 *
		 * @param[in]  pfCoords  Coordinates in play field
		 *
		 * @returns Returns coordinations in bubble field coordinations system
		 */
		SbVec3f coordsInBubbleField(const SbVec3f& pfCoords);

	public:

		/**
		 * @brief Constructor.
		 *
		 * @param[in] sceneConfig Scene configuration.
		 */
		PlayField(const SceneConfiguration &sceneConfig);

		/**
		 * @brief Destructor.
		 */
		~PlayField();

		// Override
		virtual void timeUpdate(float dt);

		// Override
		virtual void callableBackNotify(BubbleField *subject);

		// Override
		virtual void callableBackNotify(Avatar *subject);

		/**
		 * @brief Shoot a bubble.
		 */
		void fire();

		/**
		 * @brief This method is called when the bubble should be really fired
		 *
		 * This method is used to delay launch of the bubble after fire() method was
		 * called in order to allow some graphical magic.
		 */
		void launchBubble();

		/**
		 * @brief Returns half size of a single side of the play box.
		 */
		static float getSizeHalfSize() { return PLAY_BOX_SIDE_HALF_SIZE; }

		/**
		 * @brief Returns the depth of the play box.
		 */
		static float getDepth() { return PLAY_BOX_DEPTH; }

		/**
		 * @brief Returns the near bound of the back wall.
		 */
		static float getNearWallBound() { return PLAY_BOX_NEAR_WALL_BOUND; }
};

#endif
