/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    snowman_avatar.cc
 * @brief   Implementation of the snowman avatar
 * @date    2008
 *
 * Implementation for the snowman avatar.
 */

#include <cassert>
#include <string>

#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCylinder.h>

#include "snowman_avatar.h"
#include "factories/factory_avatar.h"
#include "play_field.h"

const float SnowmanAvatar::INITIAL_HAMMER_ROT = M_PI / 2.0f;

SnowmanAvatar::SnowmanAvatar():
	_rotX(0), _rotY(0), _hammerTimer(0), _timerTimeOut(0), _state(STOPPED),
	_fireDone(0)
{
	// initilize the root node
	_rootNode = new SoSeparator();

	// initialize rotations
	_rotX = new SoRotationXYZ;
	_rotX->axis = SoRotationXYZ::X;
	_rootNode->addChild(_rotX);

	_rotY = new SoRotationXYZ;
	_rotY->axis = SoRotationXYZ::Y;
	_rootNode->addChild(_rotY);

	// create the hammer
	SoSeparator* hammerSep = new SoSeparator();

	// move the hammer in the Z-axix
	SoTranslation* hammerTrans = new SoTranslation();
	hammerTrans->translation.setValue(
		SbVec3f(0.0f, 0.0f, -2.5f));
	hammerSep->addChild(hammerTrans);

	// initial rotation of the hammer
	SoRotationXYZ* hammerInitRotation = new SoRotationXYZ();
	hammerInitRotation->axis.setValue(SoRotationXYZ::X);
	hammerInitRotation->angle.setValue(-INITIAL_HAMMER_ROT);
	hammerSep->addChild(hammerInitRotation);

	_hammerTimer = new SoElapsedTime();
	_hammerTimer->speed.setValue(8.0f);
	_hammerTimer->on.setValue(false);

	// create the timer output
	_timerTimeOut = new SoSFFloat();
	_timerTimeOut->connectFrom(&(_hammerTimer->timeOut));

	// rotate the hammer
	SoRotationXYZ* _hammerRotation = new SoRotationXYZ();
	_hammerRotation->axis.setValue(SoRotationXYZ::X);
	_hammerRotation->angle.connectFrom(&(_hammerTimer->timeOut));
	hammerSep->addChild(_hammerRotation);

	// move the hammer in the Y-axix
	hammerTrans = new SoTranslation();
	hammerTrans->translation.setValue(
		SbVec3f(0.0f, PlayField::getSizeHalfSize(), 0.0f));
	hammerSep->addChild(hammerTrans);

	hammerSep->addChild(Avatar::getBubbleHammer());

	_rootNode->addChild(hammerSep);

	// move the snowman
	SoTranslation* snowmanTrans = new SoTranslation();
	snowmanTrans->translation.setValue(SbVec3f(-2.0f, 0.0f, 0.0f));
	_rootNode->addChild(snowmanTrans);

	// invert the coordinate system for the snowman
	SoScale* snowmanScale = new SoScale();
	snowmanScale->scaleFactor.setValue(SbVec3f(1.0f, 1.0f, -1.0f));
	_rootNode->addChild(snowmanScale);

	// constant rotation of the snowman
	SoRotationXYZ* rotSnowman = new SoRotationXYZ();
	rotSnowman->axis.setValue(SoRotationXYZ::Y);
	rotSnowman->angle.setValue(INITIAL_HAMMER_ROT);
	_rootNode->addChild(rotSnowman);

	// rotate the snowman with the hammer
	rotSnowman = new SoRotationXYZ();
	rotSnowman->axis.setValue(SoRotationXYZ::Y);
	rotSnowman->angle.connectFrom(&(_hammerTimer->timeOut));
	_rootNode->addChild(rotSnowman);

	// create the avatar (snowman)

	// start with snowballs

	// assign a material
	SoMaterial *material = new SoMaterial();
	//sunMat->emissiveColor.setValue(0.9f, 0.9f, 0.3f);
	material->diffuseColor.setValue(1.0f, 1.0f, 1.0f);
	material->specularColor.setValue(1.0f, 1.0f, 1.0f);
	material->shininess.setValue(1.0f);
	_rootNode->addChild(material);


	// the bottom snowball
	SoSphere* snowball = new SoSphere();
	snowball->radius.setValue(1.2f);
	_rootNode->addChild(snowball);

	// the middle snowball
	SoTranslation* trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(0.0f, 1.7f, 0.0f));
	_rootNode->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.9f);
	_rootNode->addChild(snowball);

	// the left hand
	SoSeparator* handSep = new SoSeparator();

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(1.05f, 0.0f, 0.0f));
	handSep->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.3f);
	handSep->addChild(snowball);

	_rootNode->addChild(handSep);

	// the right hand
	handSep = new SoSeparator();

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(-1.05f, 0.0f, 0.0f));
	handSep->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.3f);
	handSep->addChild(snowball);

	_rootNode->addChild(handSep);

	// the top snowball
	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(0.0f, 1.2f, 0.0f));
	_rootNode->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.5f);
	_rootNode->addChild(snowball);

	// make a nose
	SoSeparator* noseSep= new SoSeparator();

	material = new SoMaterial();
	material->diffuseColor.setValue(1.0f, 0.0f, 0.0f);
	noseSep->addChild(material);

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(0.0f, 0.0f, 0.7f));
	noseSep->addChild(trans);

	SoRotationXYZ* rot = new SoRotationXYZ();
	rot->axis.setValue(SoRotationXYZ::X);
	rot->angle.setValue(M_PI / 2.0f);
	noseSep->addChild(rot);

	SoCone* carrot = new SoCone();
	carrot->height.setValue(0.5f);
	carrot->bottomRadius.setValue(0.05f);
	noseSep->addChild(carrot);

	_rootNode->addChild(noseSep);

	// make the eyes
	SoSeparator* twoEyesSep = new SoSeparator();

	material = new SoMaterial();
	material->diffuseColor.setValue(0.0f, 0.0f, 0.0f);
	material->specularColor.setValue(1.0f, 1.0f, 1.0f);
	material->shininess.setValue(1.0f);
	twoEyesSep->addChild(material);

	// make the left eye
	SoSeparator* eyeSep= new SoSeparator();

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(0.2f, 0.1f, 0.38f));
	eyeSep->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.1f);
	eyeSep->addChild(snowball);

	twoEyesSep->addChild(eyeSep);

	// make the right eye
	eyeSep= new SoSeparator();

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(-0.2f, 0.1f, 0.38f));
	eyeSep->addChild(trans);

	snowball = new SoSphere();
	snowball->radius.setValue(0.1f);
	eyeSep->addChild(snowball);

	twoEyesSep->addChild(eyeSep);

	_rootNode->addChild(twoEyesSep);

	// make the mouth
	SoSeparator* mouthSep = new SoSeparator();

	material = new SoMaterial();
	material->diffuseColor.setValue(0.0f, 0.0f, 0.0f);
	mouthSep->addChild(material);

	trans = new SoTranslation();
	trans->translation.setValue(SbVec3f(0.0f, -0.2f, 0.45f));
	mouthSep->addChild(trans);

	rot = new SoRotationXYZ();
	rot->axis.setValue(SoRotationXYZ::Z);
	rot->angle.setValue(M_PI / 2.0f);
	mouthSep->addChild(rot);

	SoCylinder* mouth = new SoCylinder();
	mouth->radius.setValue(0.03f);
	mouth->height.setValue(0.3f);
	mouthSep->addChild(mouth);

	_rootNode->addChild(mouthSep);

	// TODO: make the hat
}

SnowmanAvatar::~SnowmanAvatar()
{
}

void SnowmanAvatar::timeUpdate(float /*dt*/)
{
	switch (_state)	{
		case STOPPED: {
				break;
			}

		case PREFIRE: {
				if (fmod(_timerTimeOut->getValue(), 2.0f * M_PI) > INITIAL_HAMMER_ROT) {
					// if hammer touches the bubble
					_state = POSTFIRE;
					_fireDone->callableBackNotify(this);
				}

				break;
			}

		case POSTFIRE: {
				if (fmod(_timerTimeOut->getValue(), 2.0f * M_PI) < INITIAL_HAMMER_ROT) {
					// if hammer touches the bubble
					_state = STOPPED;
					_hammerTimer->on.setValue(false);
				}

				break;
			}

		default:
			assert(false); // Program should never get there (invalid state)
	}
}

void SnowmanAvatar::update(Camera * subject)
{
	// Update avatar rotation according to the camera rotation
	_rotX->angle.setValue(-subject->getRotationX()->angle.getValue());
	_rotY->angle.setValue(-subject->getRotationY()->angle.getValue());
}

void SnowmanAvatar::fire(Templates::CallableBack<Avatar>* fireDone)
{
	// start the firing sequence
	// set start values
	_state = PREFIRE;
	_fireDone = fireDone;

	// initialize and unpause the hammer timer
	_hammerTimer->on.setValue(true);
}

// Functions and constants needed by Factories::FactoryAvatar
namespace
{
	// Avatar identifier for the avatars factory
	const std::string SNOWMAN_AVATAR_ID = "Snowman";

	// Function that returns an instance of the Avatar class
	Avatar *getSnowmanAvatarInstance()
	{
		return new SnowmanAvatar::SnowmanAvatar();
	}

	// Register the selected avatar at the avatars factory
	const bool SNOWMAN_AVATAR_REGISTERED = Factories::FactoryAvatar::instance().
		registerObject(SNOWMAN_AVATAR_ID, getSnowmanAvatarInstance);
}
