/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    space_environment.cc
 * @brief   SpaceEnvironment class definition.
 * @date    2008
 *
 */

#include <string>

#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/engines/SoElapsedTime.h>

#include "space_environment.h"
#include "texture_generator.h"
#include "factories/factory_environment.h"

namespace
{
	// half of the size of the SkyBox
	const float SB_SIZE_HALF = 200.0f;

	// number of vertices in the SkyBox
	const unsigned SKY_BOX_SIZE = 8;

	// number of faces of the SkyBox
	const unsigned SKY_BOX_FACES = 6;

	// number of vertices in a SkyBox face (including a '-1' ending index)
	const unsigned SKY_BOX_FACE_SIZE = 5;

	// number of vertices used for texture of a SkyBox face
	// (including a '-1' ending index)
	const unsigned SKY_BOX_TEX_SIZE = 5;

	// position of the SkyBox vertices
	// (copied from http://www.root.cz/clanky/open-inventor-vesmirna-scena-3/)
	const float SKY_BOX_VERTICES[SKY_BOX_SIZE][3] =
	{
		{-SB_SIZE_HALF, -SB_SIZE_HALF, -SB_SIZE_HALF},  // front left bottom
		{ SB_SIZE_HALF, -SB_SIZE_HALF, -SB_SIZE_HALF},  // front right bottom
		{-SB_SIZE_HALF,  SB_SIZE_HALF, -SB_SIZE_HALF},  // front left top
		{ SB_SIZE_HALF,  SB_SIZE_HALF, -SB_SIZE_HALF},  // front right top
		{-SB_SIZE_HALF, -SB_SIZE_HALF,  SB_SIZE_HALF},  // back left bottom
		{ SB_SIZE_HALF, -SB_SIZE_HALF,  SB_SIZE_HALF},  // back right bottom
		{-SB_SIZE_HALF,  SB_SIZE_HALF,  SB_SIZE_HALF},  // back left top
		{ SB_SIZE_HALF,  SB_SIZE_HALF,  SB_SIZE_HALF}   // back right top
	};

	// indices into SKY_BOX_VERTICES (-1 ends a TriangleStrip)
	const int32_t SKY_BOX_INDICES[SKY_BOX_FACES][SKY_BOX_FACE_SIZE] =
	{
		{0, 1, 2, 3, -1},   // front face
		{6, 4, 2, 0, -1},   // left face
		{5, 4, 7, 6, -1},   // back face
		{5, 7, 1, 3, -1},   // right face
		{2, 3, 6, 7, -1},   // top face
		{4, 5, 0, 1, -1}    // bottom face
	};

	// indices of the texture for each face
	const int32_t SKY_BOX_TEX_COORD_INDICES[SKY_BOX_TEX_SIZE] =
	{
		0, 1, 2, 3, -1
	};
}


SpaceEnvironment::SpaceEnvironment()
{
	// initialize the root node
	_rootNode = new SoGroup();
	SoSeparator* environ = new SoSeparator();

	// set correct light model for the background
	SoLightModel* lightModel = new SoLightModel();
	lightModel->model.setValue(SoLightModel::PHONG);
	environ->addChild(lightModel);

	// add SkyBox vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, SKY_BOX_SIZE, SKY_BOX_VERTICES);
	environ->addChild(coords);

	// array for space background textures
	for (unsigned i = 0; i < SKY_BOX_FACES; ++i) {
		// add the SkyBox texture
		// TODO: dispose of magic constants & check for valid return value
		SoTexture2* tex = TextureGenerator::generateSpaceBackground(512, 512, 500);
		tex->model.setValue(SoTexture2::DECAL);  // discard lighting result
		environ->addChild(tex);

		// create the SkyBox walls
		SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
		strip->coordIndex.setValues(0, SKY_BOX_FACE_SIZE, SKY_BOX_INDICES[i]);
		strip->textureCoordIndex.setValues(0, SKY_BOX_TEX_SIZE, SKY_BOX_TEX_COORD_INDICES);
		environ->addChild(strip);
	}

	_rootNode->addChild(environ);

	// create the Earth
	SoSeparator* earthSep = new SoSeparator();

	SoElapsedTime* earthElapsedTime = new SoElapsedTime();
	earthElapsedTime->speed.setValue(0.08f);

	SoElapsedTime* moonElapsedTime = new SoElapsedTime();
	moonElapsedTime->speed.setValue(0.02f);

	// rotate the Moon around Earth
	SoRotationXYZ* yRot = new SoRotationXYZ();
	yRot->axis.setValue(SoRotationXYZ::Z);
	yRot->angle.setValue(-M_PI / 4.0f);
	earthSep->addChild(yRot);

	// insert a light
	SoDirectionalLight* earthLight = new SoDirectionalLight();
	earthLight->direction.setValue(SbVec3f(0.5f, 0.0f, 4.0f));
	earthSep->addChild(earthLight);

	// move the Earth to its position
	SoTranslation* earthTrans = new SoTranslation();
	earthTrans->translation.setValue(SbVec3f(45.0f, -20.0f, 35.0f));
	earthSep->addChild(earthTrans);

	// create the Moon
	SoSeparator* moonSep = new SoSeparator();

	// set the Moon texture
	moonSep->addChild(TextureGenerator::generateMoonSurface());

	// rotate the Moon around Earth
	SoRotationXYZ* moonRot = new SoRotationXYZ();
	moonRot->axis.setValue(SoRotationXYZ::Y);
	moonRot->angle.connectFrom(&(moonElapsedTime->timeOut));
	moonSep->addChild(moonRot);

	// move the Moon to its position
	SoTranslation* moonTrans = new SoTranslation();
	moonTrans->translation.setValue(SbVec3f(-120.0f, 0.0f, 0.0f));
	moonSep->addChild(moonTrans);

	//// rotate the Moon around its axis
	//moonRot = new SoRotationXYZ();
	//moonRot->axis.setValue(SoRotationXYZ::Y);
	//moonRot->angle.connectFrom(&(moonElapsedTime->timeOut));
	//moonSep->addChild(moonRot);

	// create the Moon
	SoSphere* moonSphere = new SoSphere();
	moonSphere->radius.setValue(14.0f);
	moonSep->addChild(moonSphere);

	earthSep->addChild(moonSep);

	// set the Earth texture
	earthSep->addChild(TextureGenerator::generateEarthSurface());

	// rotate the Earth
	SoRotationXYZ* earthRot = new SoRotationXYZ();
	earthRot->axis.setValue(SoRotationXYZ::Y);
	earthRot->angle.connectFrom(&(earthElapsedTime->timeOut));
	earthSep->addChild(earthRot);

	// the complexity of the Earth
	SoComplexity* earthComplexity = new SoComplexity();
	earthComplexity->value.setValue(2.0f);
	earthComplexity->textureQuality.setValue(1.0f);
	earthSep->addChild(earthComplexity);

	// create the Earth
	SoSphere* earthSphere = new SoSphere();
	earthSphere->radius.setValue(38.0f);
	earthSep->addChild(earthSphere);

	_rootNode->addChild(earthSep);

	// create lights
	SoPointLight* light = new SoPointLight();
	light->location.setValue(SbVec3f(-10.0f, -10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);

	light = new SoPointLight();
	light->location.setValue(SbVec3f(10.0f, 10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);
}

SpaceEnvironment::~SpaceEnvironment()
{
}

void SpaceEnvironment::timeUpdate(float /* dt */)
{
	// do nothing
}

// Functions and constants needed by Factories::FactoryEnvironment
namespace
{
	// SpaceEnvironment identifier for the environments factory
	const std::string SPACE_ENVIRONMENT_ID = "Space";

	// Function that returns an instance of the SpaceEnvironment class
	Environment *getSpaceEnvironmentInstance()
	{
		return new SpaceEnvironment::SpaceEnvironment();
	}

	// Register the selected environment at the environments factory
	const bool SPACE_ENVIRONMENT_REGISTERED = Factories::FactoryEnvironment::instance().
		registerObject(SPACE_ENVIRONMENT_ID, getSpaceEnvironmentInstance);
}
