/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    factory.h
 * @brief   Implementation of a factory that creates instances of objects.
 * @date    2008
 *
 * Part of the code taken and modified from Loki library presented by:
 * Alexandrescu, Andrei. "Modern C++ Design: Generic Programming and Design
 *     Patterns Applied". Copyright (c) 2001. Addison-Wesley.
 * Copyright (c) 2001 by Andrei Alexandrescu
 * Permission to use, copy, modify, distribute and sell this software for any
 *     purpose is hereby granted without fee, provided that the above copyright
 *     notice appear in all copies and that both that copyright notice and this
 *     permission notice appear in supporting documentation.
 */

#ifndef _TEMPLATES__FACTORY_H_
#define _TEMPLATES__FACTORY_H_

#include <vector>
#include <map>
#include <cassert>

/**
 * @brief All project templates.
 */
namespace Templates
{
	template <typename, class>
	class DefaultFactoryError;

	template <class, typename, typename, template<typename, class> class>
	class Factory;
}

/**
 * @brief Manages the "Unknown Type" error in an object factory.
 */
template <typename IdentifierType, class AbstractProduct>
class Templates::DefaultFactoryError
{
	protected:

		/**
		 * @brief Default constructor.
		 */
		DefaultFactoryError() { }

		/**
		 * @brief Destructor.
		 */
		~DefaultFactoryError() { }

	public:

		/**
		 * @brief Reaction to the "Unknown Type" error.
		 *
		 * @param[in] id Selected object ID.
		 *
		 * This default implementation uses only an assert and returns
		 * a null pointer.
		 */
		static AbstractProduct *onUnknownType(IdentifierType /*id*/)
		{
			// Trying to instantiate an unknown object!
			assert(false);
			return 0;
		}
};

/**
 * @brief Implementation of a generic object factory.
 *
 * Implements "object factory" design pattern - creates instances of objects
 * according to the given object identifier.
 */
template <
	class AbstractProduct,
	typename IdentifierType,
	typename ProductCreator = AbstractProduct *(*)(),
	template<typename, class>
		class FactoryErrorPolicy = Templates::DefaultFactoryError
>
class Templates::Factory: public FactoryErrorPolicy<IdentifierType, AbstractProduct>
{
	private:

		/// Type of container used to store object IDs.
		typedef std::map<IdentifierType, ProductCreator> IdToProductMap;

		/// Container used to store object IDs.
		IdToProductMap _associations;

	public:

		/**
		 * @brief Default constructor.
		 */
		Factory(): _associations() { }

		/**
		 * @brief Registers the selected object.
		 *
		 * @param[in] id      Selected object ID.
		 * @param[in] creator Creator that will be used to create instances.
		 * @return            True if the registration was successfull,
		 *                    false otherwise.
		 *
		 * Each ID can be registered only once (if someone tries to register
		 * the same object twice, this function will return false).
		 */
		bool registerObject(IdentifierType id, ProductCreator creator)
		{
			return _associations.insert(
				typename IdToProductMap::value_type(id, creator)).second;
		}

		/**
		 * @brief Unregisters the selected object.
		 *
		 * @param[in] id Selected object ID.
		 * @return       True if the registration was successfull,
		 *               false otherwise.
		 */
		bool unregisterObject(const IdentifierType & id) const
		{
			return _associations.erase(id) == 1;
		}

		/**
		 * @brief Creates an instance of the selected object.
		 *
		 * @param[in] id Selected object ID.
		 * @return       Instance of the selected object.
		 *
		 * Object instance is allocated on the heap, so you have to delete
		 * the allocated memory manually.
		 */
		AbstractProduct *createObject(const IdentifierType & id) const
		{
			typename IdToProductMap::const_iterator i = _associations.find(id);
			if (i != _associations.end()) {
				// Instantiate the object via registered function
				return (i->second)();
			}

			return onUnknownType(id);
		}

		/**
		 * @brief Returns a vector of all registered object IDs.
		 */
		std::vector<IdentifierType> registeredObjects() const
		{
			std::vector<IdentifierType> regObjects;
			for (typename IdToProductMap::const_iterator i = _associations.begin();
				i != _associations.end(); ++i) {
				regObjects.push_back(i->first);
			}
			return regObjects;
		}
};

#endif
