/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    observable.h
 * @brief   Implementation of a generic "typed" observer (Observable class).
 * @date    2008
 *
 * Implementation is based on http://kelpi.com/script/d73d5b.
 */

#ifndef _TEMPLATES__OBSERVABLE_H_
#define _TEMPLATES__OBSERVABLE_H_

#include <iterator>
#include <algorithm>
#include <deque>

#include "observer.h"

/**
 * @brief All project templates.
 */
namespace Templates
{
	template <class T>
	class Observable;
}

/**
 * @brief Implementation of a generic "typed" observer (observable part).
 *
 * Implements "observer" design pattern.
 *
 * Usage:
 *
 * @code
 * class View: public Observable<View>
 * {
 *     private:
 *         std::string _name;
 *     public:
 *         View(std::string name): Observable<View>(this), _name(name) {}
 *         ~View() {}
 *         std::string getName() {return _name;}
 * };
 *
 * @endcode
 *
 * @see Observer
 */
template <class T>
class Templates::Observable
{
	private:

		/// Actual object that will be passed to Observer::update() calls.
		T *_self;

		/// Type of a container to store observers.
		typedef std::deque<Observer<T> *> ObserverContainer;

		/// Stored observers.
		ObserverContainer _observers;

	public:
		/**
		 * @brief Constructor.
		 *
		 * @param[in] self Object that will be passed to Observer::update() calls.
		 */
		Observable(T *self): _self(self), _observers() { }

		/**
		 * @brief Destructor.
		 */
		virtual ~Observable() { }

		/**
		 * @brief Adds a new observer to the list of observers.
		 *
		 * @param[in] observer Observer to be added.
		 */
		void addObserver(Observer<T> *observer)
		{
			_observers.push_back(observer);
		}

		/**
		 * @brief Deletes selected observer from the list of observers.
		 *
		 * @param[in] observer Observer to be deleted.
		 */
		void deleteObserver(Observer<T> *observer)
		{
			_observers.erase(std::remove_if(_observers.begin(), _observers.end(),
				std::bind2nd(std::equal_to<Observer<T> *>(), observer)));
		}

		/**
		 * @brief Deletes all observers.
		 */
		void cleanObservers()
		{
			_observers.clean();
		}

		/**
		 * @brief Notifies all stored observers by calling Observer::update() on them.
		 *
		 * @see Observer::update()
		 */
		void notifyObservers()
		{
			for (typename ObserverContainer::iterator iter = _observers.begin();
				iter != _observers.end(); ++iter) {
				(*iter)->update(_self);
			}
		}
};

#endif
