/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    texture_generator.cc
 * @brief   Implementation of the TextureGenerator class
 * @date    2008
 *
 * Implementation of the TextureGenerator class; a static class for generation
 * of various textures.
 */

#include <cstdlib>
#include <iostream>

#include "texture_generator.h"
#include "path_functions.h"
#include "convert.h"

const std::string TextureGenerator::PATH_TO_ALBUM_DIR = "imgs/albums";
const std::string TextureGenerator::ALBUM_IMG_PREFIX  = "image";
const std::string TextureGenerator::ALBUM_IMG_SUFFIX  = ".jpg";

const std::string TextureGenerator::PATH_TO_MODEL_DIR = "models";
const std::string TextureGenerator::MODEL_SUFFIX      = ".iv";

SoTexture2* TextureGenerator::generateSpaceBackground(
	unsigned width, unsigned height, unsigned starCount)
{
	SoTexture2* tex = new SoTexture2();

	// prepare the properties of the image
	unsigned long bufSize = static_cast<unsigned long>(width) * height;
	SbVec2s imgSize(width, height);
	int componentsNumber = NC_RGBA;

	// set the size and pixel format of the image; this will initialize the
	// background to black colour
	tex->image.setValue(imgSize, componentsNumber, 0);

	// start editing the pixel buffer
	unsigned char* buf = tex->image.startEditing(imgSize, componentsNumber);

	// clear the buffer
	for (unsigned long i = 0; i < bufSize; ++i) {
		setRGBAPixel(buf, bufSize, i, 0x00, 0x00, 0x00, 0xFF);
	}

	// generate stars
	for (unsigned i = 0; i < starCount; ++i) {
		unsigned long x = (std::rand() % (width - 2)) + 1;
		unsigned long y = (std::rand() % (height - 2)) + 1;

		unsigned char red   = 0xFF - (std::rand() % SPACE_STAR_LIGHT_MAX_DEVIATION);
		unsigned char green = 0xFF - (std::rand() % SPACE_STAR_LIGHT_MAX_DEVIATION);
		unsigned char blue  = 0xFF - (std::rand() % SPACE_STAR_LIGHT_MAX_DEVIATION);
		unsigned char alpha	= 0xFF;

		// draw a point
		setRGBAPixel(buf, bufSize, cartesianToLinear(width, x, y),
			red, green, blue, alpha);
	}

	// flush the changes of the pixel buffer
	tex->image.finishEditing();

	return tex;
}

SoTexture2* TextureGenerator::generateEarthSurface()
{
	SoTexture2* tex = new SoTexture2();
	tex->filename.setValue(PathFunctions::createProperFilePath(
		"imgs/earth_lowres.jpg").c_str());

	return tex;
}

SoTexture2* TextureGenerator::generateMoonSurface()
{
	SoTexture2* tex = new SoTexture2();
	tex->filename.setValue(PathFunctions::createProperFilePath(
		"imgs/moon.jpg").c_str());

	return tex;
}

SoTexture2* TextureGenerator::generateWireWall()
{
	SoTexture2* tex = new SoTexture2();
	tex->filename.setValue(PathFunctions::createProperFilePath(
		"imgs/smoke.jpg").c_str());

	return tex;
}

SoTexture2* TextureGenerator::generateTunnelWall(SbColor4f& startColour,
	SbColor4f& endColour, unsigned lengthHalf)
{
	float redFloat, greenFloat, blueFloat, alphaFloat;

	// get the start colour components
	startColour.getValue(redFloat, greenFloat, blueFloat, alphaFloat);
	unsigned char startRed = colourFloatToUC(redFloat);
	unsigned char startGreen = colourFloatToUC(greenFloat);
	unsigned char startBlue = colourFloatToUC(blueFloat);
	unsigned char startAlpha = colourFloatToUC(alphaFloat);

	// get the end colour components
	endColour.getValue(redFloat, greenFloat, blueFloat, alphaFloat);
	unsigned char endRed = colourFloatToUC(redFloat);
	unsigned char endGreen = colourFloatToUC(greenFloat);
	unsigned char endBlue = colourFloatToUC(blueFloat);
	unsigned char endAlpha = colourFloatToUC(alphaFloat);

	// compute the difference between the start and the end components
	int diffRed = static_cast<int>(endRed) - startRed;
	int diffGreen = static_cast<int>(endGreen) - startGreen;
	int diffBlue = static_cast<int>(endBlue) - startBlue;
	int diffAlpha = static_cast<int>(endAlpha) - startAlpha;

	// create the texture
	SoTexture2* tex = new SoTexture2();

	// prepare the properties of the image
	unsigned long bufSize = static_cast<unsigned long>(2 * lengthHalf);
	SbVec2s imgSize(bufSize, 1);
	int componentsNumber = NC_RGBA;

	// set the size and pixel format of the image; this will initialize the
	// background to black colour
	tex->image.setValue(imgSize, componentsNumber, 0);

	// start editing the pixel buffer
	unsigned char* buf = tex->image.startEditing(imgSize, componentsNumber);

	for (unsigned i = 0; i < lengthHalf; ++i) {
		unsigned char red   = startRed + ((i * diffRed) / lengthHalf);
		unsigned char green = startGreen + ((i * diffGreen) / lengthHalf);
		unsigned char blue  = startBlue + ((i * diffBlue) / lengthHalf);
		unsigned char alpha	= startAlpha + ((i * diffAlpha) / lengthHalf);

		// draw a point
		setRGBAPixel(buf, bufSize, i, red, green, blue, alpha);
	}

	for (unsigned i = 0; i < lengthHalf; ++i) {
		unsigned char red   = endRed - ((i * diffRed) / lengthHalf);
		unsigned char green = endGreen - ((i * diffGreen) / lengthHalf);
		unsigned char blue  = endBlue - ((i * diffBlue) / lengthHalf);
		unsigned char alpha	= endAlpha - ((i * diffAlpha) / lengthHalf);

		// draw a point
		setRGBAPixel(buf, bufSize, lengthHalf + i, red, green, blue, alpha);
	}

	// flush the changes of the pixel buffer
	tex->image.finishEditing();

	return tex;
}

void TextureGenerator::setRGBAPixel(unsigned char* pixelBuffer,
	unsigned long bufSize, unsigned long position,
	unsigned char red, unsigned char green, unsigned char blue,
	unsigned char alpha)
{
	// check validity of input parameters
	assert(position < bufSize);

	// set all colour components of the pixel
	pixelBuffer[TextureGenerator::NC_RGBA * position + 0] = red;
	pixelBuffer[TextureGenerator::NC_RGBA * position + 1] = green;
	pixelBuffer[TextureGenerator::NC_RGBA * position + 2] = blue;
	pixelBuffer[TextureGenerator::NC_RGBA * position + 3] = alpha;
}

AlbumContainer TextureGenerator::generateAlbum(const std::string& albumName)
{
	AlbumContainer cont;

	std::string prefix = PATH_TO_ALBUM_DIR + "/" + albumName + "/" +
		ALBUM_IMG_PREFIX;

	std::string firstExistingFile;

	for (unsigned i = 0; i < 6; ++i) {
		SoTexture2* tex = new SoTexture2();

		std::string textureFile = PathFunctions::createProperFilePath(
			prefix + Convert::UnsignedToString(i) + ALBUM_IMG_SUFFIX);
		// if the file doesn't exist, use the first file that exists instead
		if (PathFunctions::fileExists(textureFile)) {
			tex->filename.setValue(textureFile.c_str());
			if (firstExistingFile.empty()) {
				firstExistingFile = textureFile;
			}
		} else if (!firstExistingFile.empty()) {
			tex->filename.setValue(firstExistingFile.c_str());
		}

		cont.push_back(tex);
	}

	return cont;
}

SoFile* TextureGenerator::generateModel(const std::string& modelName)
{
	SoFile* file = new SoFile();
	file->name.setValue(PathFunctions::createProperFilePath(PATH_TO_MODEL_DIR + "/" +
		modelName + MODEL_SUFFIX).c_str());

	return file;
}

unsigned char TextureGenerator::colourFloatToUC(float colourValue)
{
	return static_cast<unsigned char>(round(255 * colourValue));
}
