/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    texture_generator.h
 * @brief   Header file for the TextureGenerator class
 * @date    2008
 *
 * Header file for the TextureGenerator class; a static class for generation
 * of various textures.
 */

#ifndef _TEXTURE_GENERATOR_H_
#define _TEXTURE_GENERATOR_H_

#include <vector>

#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoFile.h>
#include <Inventor/SbColor4f.h>

/// Container for album textures
typedef std::vector<SoTexture2*> AlbumContainer;

/**
 * @brief   A static class for generation of various textures
 *
 * Comprises several class methods for generation of various textures.
 *
 * Uses "library class" (aka "static helper") design pattern.
 */
class TextureGenerator
{
	private:

		/// Number of components for an RGBA pixel
		static const int NC_RGBA = 4;

		/// Maximum deviation of every colour component of a star light in the
		/// space background from 0xFF
		static const unsigned char SPACE_STAR_LIGHT_MAX_DEVIATION = 0x10;

	public:

		/// Path to the user album directory
		static const std::string PATH_TO_ALBUM_DIR;

		/// Prefix of user album images
		static const std::string ALBUM_IMG_PREFIX;

		/// Suffix of user album images
		static const std::string ALBUM_IMG_SUFFIX;

		/// Path to the user model directory
		static const std::string PATH_TO_MODEL_DIR;

		/// Suffix of user models
		static const std::string MODEL_SUFFIX;

	private:

		// Disable both constructors, destructor and assignment operator
		TextureGenerator();
		~TextureGenerator();
		TextureGenerator(const TextureGenerator &);
		TextureGenerator & operator=(const TextureGenerator &);

		/**
		 * @brief  A static helper method for setting a pixel of an RGBA pixel
		 *         buffer to desired colour.
		 *
		 * A static helper method that sets a pixel at given position (determined
		 * as a linear position) in an RGBA pixel buffer to desired colour.
		 *
		 * @param[in]  pixelBuffer  Pointer to the start of the pixel buffer
		 * @param[in]  bufSize      Size of the buffer to enable control for
		 *                          out-of-bounds access
		 * @param[in]  position     Linear position of the pixel in the pixel
		 *                          buffer
		 * @param[in]  red          Red component of the colour
		 * @param[in]  green        Green component of the colour
		 * @param[in]  blue         Blue component of the colour
		 * @param[in]  alpha        Alpha component of the colour
		 */
		static void setRGBAPixel(unsigned char* pixelBuffer,
			unsigned long bufSize, unsigned long position,
			unsigned char red, unsigned char green, unsigned char blue,
			unsigned char alpha);

		/**
		 * @brief  A static helper method for conversion of Cartesian coordinates
		 *         to linear coordinates into a buffer
		 *
		 * A static helper method that converts Cartesian coordinates to linear
		 * coordinates into a buffer. The width of the buffer needs to be known.
		 *
		 * @param[in]  width  Width of the buffer
		 * @param[in]  x      X-coordinate into the buffer
		 * @param[in]  y      Y-coordinate into the buffer
		 * @return     Linear address in the buffer
		 */
		static unsigned long cartesianToLinear(unsigned long width,
			unsigned long x, unsigned long y)
		{
			return x + y*width;
		}

		/**
		 * @brief  A static helper method for conversion of float representation
		 *         of colour components to unsigned char
		 *
		 * A static helper method that converts float colour components <0.0; 1.0>
		 * to their unsigned char representation <0; 255>.
		 *
		 * @param[in]  colourValue  The float value of the colour component
		 * @return     Unsinged char value of the colour component
		 */
		static unsigned char colourFloatToUC(float colourValue);

	public:

		/**
		 * @brief  Static method for generation of a space background texture
		 *
		 * @param[in]  width      The width of the texture
		 * @param[in]  height     The height of the texture
		 * @param[in]  starCount  The number of stars in the texture
		 * @return     Generated texture
		 */
		static SoTexture2* generateSpaceBackground(
			unsigned width, unsigned height, unsigned starCount);

		/**
		 * @brief  Static method for generation of an Earth surface texture 
		 *
		 * @return     Generated texture
		 */
		static SoTexture2* generateEarthSurface();

		/**
		 * @brief  Static method for generation of a Moon surface texture 
		 *
		 * @return     Generated texture
		 */
		static SoTexture2* generateMoonSurface();

		/**
		 * @brief  Static method for generation of a network wall texture
		 *
		 * @return     Generated texture
		 */
		static SoTexture2* generateWireWall();

		/**
		 * @brief  Static method for generation of tunnel wall texture
		 *
		 * This method generates a texture for the tunnel wall used in
		 * TunnelPlayBoxGraphics. The method accepts two colours -> the texture
		 * will be generated as a gradient of colours between these two. The size
		 * of the texture will be length x 1 px.
		 *
		 * @param[in]  startColour  The start colour of the tunnel wall
		 * @param[in]  endColour    The end colour of the tunnel wall
		 * @param[in]  lengthHalf   Half the length of the texture
		 * @return     Generated texture
		 */
		static SoTexture2* generateTunnelWall(SbColor4f& startColour,
			SbColor4f& endColour, unsigned lengthHalf);

		/**
		 * @brief  Static method for loading of album images
		 *
		 * This static method loads images from album of given name and returns
		 * them as textures in a container.
		 *
		 * @param[in]  albumName  The name of the album (the name of the album
		 *                        directory in the filesystem needs to be the same)
		 * @return     Generated album textures
		 */
		static AlbumContainer generateAlbum(const std::string& albumName);

		/**
		 * @brief  Static method for loading of models
		 *
		 * This static method loads a model of given name
		 *
		 * @param[in]  modelName  The name of the model (the name of the model
		 *                        file in the filesystem needs to be the same)
		 * @return     Generated model
		 */
		static SoFile* generateModel(const std::string& modelName);
};

#endif
