/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    tunnel_play_box_graphics.cc
 * @brief   TunnelPlayBoxGraphics class definition.
 * @date    2008
 *
 */

#include <string>

#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoTransparencyType.h>

#include "tunnel_play_box_graphics.h"
#include "texture_generator.h"
#include "play_field.h"
#include "factories/factory_play_box_graphics.h"


const SbColor4f TunnelPlayBoxGraphics::TUNNEL_COLOUR_START(1.0f, 0.0f, 0.0f, 1.0f);
const SbColor4f TunnelPlayBoxGraphics::TUNNEL_COLOUR_END(0.0f, 0.0f, 1.0f, 0.0f);

namespace
{
	// Half the length of the tunnel texture
	const unsigned TUNNEL_TEX_LENGTH_HALF = 512;

	// Half of the side of the play box
	const float BOX_SIDE_HALF = PlayField::getSizeHalfSize();

	// Depth of the play box
	const float BOX_DEPTH = PlayField::getDepth();

	// Transparency of the rear wall
	const float TRANSPARENCY_TRANSPARENT = 0.8f;

	// number of vertices in the tunnel play field
	const unsigned TPF_SIZE = 8;

	// number of faces of the tunnel play field
	const unsigned TPF_FACES = 4;

	// number of vertices in a tunnel play field face
	// (including a '-1' ending index)
	const unsigned TPF_FACE_SIZE = 5;

	// number of lines in the play box
	const unsigned TPF_LINES = 4;

	// number of vertices used for texture of a tunnel play field face
	// (including a '-1' ending index)
	const unsigned TPF_LINE_SIZE = 3;

	// position of the tunnel play field vertices
	const float TPF_VERTICES[TPF_SIZE][3] =
	{
		{-BOX_SIDE_HALF, -BOX_SIDE_HALF,         0},  // front right bottom
		{ BOX_SIDE_HALF, -BOX_SIDE_HALF,         0},  // front left bottom
		{-BOX_SIDE_HALF,  BOX_SIDE_HALF,         0},  // front right top
		{ BOX_SIDE_HALF,  BOX_SIDE_HALF,         0},  // front left top
		{-BOX_SIDE_HALF, -BOX_SIDE_HALF, BOX_DEPTH},  // back right bottom
		{ BOX_SIDE_HALF, -BOX_SIDE_HALF, BOX_DEPTH},  // back left bottom
		{-BOX_SIDE_HALF,  BOX_SIDE_HALF, BOX_DEPTH},  // back right top
		{ BOX_SIDE_HALF,  BOX_SIDE_HALF, BOX_DEPTH}   // back left top
	};

	// indices into TPF_VERTICES (-1 ends a TriangleStrip)
	const int32_t TPF_INDICES[TPF_FACES][TPF_FACE_SIZE] =
	{
		{6, 4, 2, 0, -1},   // right face
		{5, 7, 1, 3, -1},   // left face
		{7, 6, 3, 2, -1},   // top face
		{4, 5, 0, 1, -1}    // bottom face
	};

	// indices of the back wall into TPF_VERTICES (-1 ends a TriangleStrip)
	const int32_t TPF_BWALL_INDICES[TPF_FACE_SIZE] =
	{
		0, 2, 1, 3, -1
	};

	// indices of line vertices (from TPF_VERTICS) for the tunnel play box
	// (-1 ends a LineStrip)
	const int32_t TPF_LINE_INDICES[TPF_LINES][TPF_LINE_SIZE] =
	{
		{0, 4, -1},
		{1, 5, -1},
		{2, 6, -1},
		{3, 7, -1}
	};
}


TunnelPlayBoxGraphics::TunnelPlayBoxGraphics() :
	_textureCoord(0), _textureOffset(0.0f)
{
	_rootNode = new SoSeparator();

	// create the walls of the box
	_rootNode->addChild(createWalls());

	// create the back wall of the box
	_rootNode->addChild(createBackWall());
}

SoGroup* TunnelPlayBoxGraphics::createWalls()
{
	SoSeparator* boxSep = new SoSeparator();

	// set correct light model for the box
	SoLightModel* lightModel = new SoLightModel();
	lightModel->model.setValue(SoLightModel::BASE_COLOR);
	boxSep->addChild(lightModel);

	// wall texture
	SbColor4f startColour = TUNNEL_COLOUR_START;
	SbColor4f endColour = TUNNEL_COLOUR_END;
	boxSep->addChild(TextureGenerator::generateTunnelWall(
		startColour, endColour, TUNNEL_TEX_LENGTH_HALF));

	// add play field vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, TPF_SIZE, TPF_VERTICES);
	boxSep->addChild(coords);

	_textureCoord = new SoTextureCoordinate2();
	_textureCoord->point.set1Value(0, SbVec2f(1.0f, 0.00f));
	_textureCoord->point.set1Value(1, SbVec2f(1.0f, 0.75f));
	_textureCoord->point.set1Value(2, SbVec2f(1.0f, 0.25f));
	_textureCoord->point.set1Value(3, SbVec2f(1.0f, 0.50f));
	_textureCoord->point.set1Value(4, SbVec2f(0.0f, 0.00f));
	_textureCoord->point.set1Value(5, SbVec2f(0.0f, 0.75f));
	_textureCoord->point.set1Value(6, SbVec2f(0.0f, 0.25f));
	_textureCoord->point.set1Value(7, SbVec2f(0.0f, 0.50f));
	boxSep->addChild(_textureCoord);

	// create walls of the play field
	SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
	// send the whole field of indices to the triangle strip
	strip->coordIndex.setValues(0, TPF_FACES * TPF_FACE_SIZE, *TPF_INDICES);
	boxSep->addChild(strip);

	return boxSep;
}


SoGroup* TunnelPlayBoxGraphics::createBackWall()
{
	// draw the back wall
	SoSeparator* backWallSep = new SoSeparator();

	// add play field vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, TPF_SIZE, TPF_VERTICES);
	backWallSep->addChild(coords);

	SoMaterial* material = new SoMaterial();
	material->diffuseColor.setValue(0.5f, 0.5f, 0.0f);
	material->transparency.setValue(TRANSPARENCY_TRANSPARENT);
	backWallSep->addChild(material);

	backWallSep->addChild(_wallTranslation);

	// create the back wall of the play field
	SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
	// send the whole field of indices to the triangle strip
	strip->coordIndex.setValues(0, TPF_FACE_SIZE, TPF_BWALL_INDICES);
	backWallSep->addChild(strip);

	return backWallSep;
}


TunnelPlayBoxGraphics::~TunnelPlayBoxGraphics()
{
}

void TunnelPlayBoxGraphics::timeUpdate(float dt)
{
	_textureOffset += dt;

	if (_textureOffset <= -1.0f) {
		_textureOffset = fmod(_textureOffset, 1.0f);
	}

	// move the texture on the walls
	_textureCoord->point.set1Value(0, SbVec2f(1.0f - _textureOffset, 0.00f));
	_textureCoord->point.set1Value(1, SbVec2f(1.0f - _textureOffset, 0.75f));
	_textureCoord->point.set1Value(2, SbVec2f(1.0f - _textureOffset, 0.25f));
	_textureCoord->point.set1Value(3, SbVec2f(1.0f - _textureOffset, 0.50f));
	_textureCoord->point.set1Value(4, SbVec2f(0.0f - _textureOffset, 0.00f));
	_textureCoord->point.set1Value(5, SbVec2f(0.0f - _textureOffset, 0.75f));
	_textureCoord->point.set1Value(6, SbVec2f(0.0f - _textureOffset, 0.25f));
	_textureCoord->point.set1Value(7, SbVec2f(0.0f - _textureOffset, 0.50f));
}

// Functions and constants needed by Factories::FactoryPlayBoxGraphics
namespace
{
	// TunnelPlayBoxGraphics identifier for the play box graphics factory
	const std::string TUNNEL_PLAY_BOX_GRAPHICS_ID = "Tunnel";

	// Function that returns an instance of the TunnelPlayBoxGraphics class
	PlayBoxGraphics *getTunnelPlayBoxGraphicsInstance()
	{
		return new TunnelPlayBoxGraphics::TunnelPlayBoxGraphics();
	}

	// Register the selected play box graphics at the play box graphics factory
	const bool TUNNEL_PLAY_BOX_GRAPHICS_REGISTERED = Factories::FactoryPlayBoxGraphics::instance().
		registerObject(TUNNEL_PLAY_BOX_GRAPHICS_ID, getTunnelPlayBoxGraphicsInstance);
}

