/**
 *  Project:  PGR 2008
 *  Authors:  Ondrej Lengal,   xlenga00@stud.fit.vutbr.cz
 *            Libor Polcak,    xpolca03@stud.fit.vutbr.cz
 *            Boris Prochazka, xproch63@stud.fit.vutbr.cz
 *            Petr Zemek,      xzemek02@stud.fit.vutbr.cz
 *
 * @file    wire_environment.cc
 * @brief   WireEnvironment class definition.
 * @date    2008
 *
 */

#include <cmath>
#include <cstdlib>
#include <string>

#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTranslation.h>

#include "wire_environment.h"
#include "texture_generator.h"
#include "factories/factory_environment.h"

namespace
{
	// half of the size of the SkyBox
	const float SB_SIZE_HALF = 400.0f;

	// half of the size of the wire
	const float WIRE_SIZE_HALF = SB_SIZE_HALF * 2;

	// number of generated wires around the play box
	const int WIRES_COUNT = 30;

	// number of vertices in the SkyBox
	const unsigned SKY_BOX_SIZE = 8;

	// number of faces of the SkyBox
	const unsigned SKY_BOX_FACES = 6;

	// number of vertices in a SkyBox face (including a '-1' ending index)
	const unsigned SKY_BOX_FACE_SIZE = 5;

	// number of vertices used for texture of a SkyBox face
	// (including a '-1' ending index)
	const unsigned SKY_BOX_TEX_SIZE = 5;

	// position of the SkyBox vertices
	// (copied from http://www.root.cz/clanky/open-inventor-vesmirna-scena-3/)
	const float SKY_BOX_VERTICES[SKY_BOX_SIZE][3] =
	{
		{-SB_SIZE_HALF, -SB_SIZE_HALF, -SB_SIZE_HALF},  // front left bottom
		{ SB_SIZE_HALF, -SB_SIZE_HALF, -SB_SIZE_HALF},  // front right bottom
		{-SB_SIZE_HALF,  SB_SIZE_HALF, -SB_SIZE_HALF},  // front left top
		{ SB_SIZE_HALF,  SB_SIZE_HALF, -SB_SIZE_HALF},  // front right top
		{-SB_SIZE_HALF, -SB_SIZE_HALF,  SB_SIZE_HALF},  // back left bottom
		{ SB_SIZE_HALF, -SB_SIZE_HALF,  SB_SIZE_HALF},  // back right bottom
		{-SB_SIZE_HALF,  SB_SIZE_HALF,  SB_SIZE_HALF},  // back left top
		{ SB_SIZE_HALF,  SB_SIZE_HALF,  SB_SIZE_HALF}   // back right top
	};

	// indices into SKY_BOX_VERTICES (-1 ends a TriangleStrip)
	const int32_t SKY_BOX_INDICES[SKY_BOX_FACES][SKY_BOX_FACE_SIZE] =
	{
		{0, 1, 2, 3, -1},   // front face
		{6, 4, 2, 0, -1},   // left face
		{5, 4, 7, 6, -1},   // back face
		{5, 7, 1, 3, -1},   // right face
		{2, 3, 6, 7, -1},   // top face
		{4, 5, 0, 1, -1}    // bottom face
	};

	// indices of the texture for each face
	const int32_t SKY_BOX_TEX_COORD_INDICES[SKY_BOX_TEX_SIZE] =
	{
		0, 1, 2, 3, -1
	};
}


WireEnvironment::WireEnvironment()
{
	// initialize the root node
	_rootNode = new SoGroup();
	SoSeparator* environ = new SoSeparator();

	// set correct light model for the background
	SoLightModel* lightModel = new SoLightModel();
	lightModel->model.setValue(SoLightModel::PHONG);
	environ->addChild(lightModel);

	// create lights
	SoPointLight* light = new SoPointLight();
	light->location.setValue(SbVec3f(-10.0f, -10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);

	// add SkyBox vertices
	SoCoordinate3* coords = new SoCoordinate3();
	coords->point.setValues(0, SKY_BOX_SIZE, SKY_BOX_VERTICES);
	environ->addChild(coords);

	// array for Wire background
	for (unsigned i = 0; i < SKY_BOX_FACES; ++i) {
		// add the SkyBox texture
		// TODO: dispose of magic constants & check for valid return value
		SoTexture2* tex = TextureGenerator::generateWireWall();
		tex->model.setValue(SoTexture2::DECAL);  // discard lighting result
		environ->addChild(tex);

		// create the SkyBox walls
		SoIndexedTriangleStripSet* strip = new SoIndexedTriangleStripSet();
		strip->coordIndex.setValues(0, SKY_BOX_FACE_SIZE, SKY_BOX_INDICES[i]);
		strip->textureCoordIndex.setValues(0, SKY_BOX_TEX_SIZE, SKY_BOX_TEX_COORD_INDICES);
		environ->addChild(strip);
	}

	_rootNode->addChild(environ);

	// Create light
	light = new SoPointLight();
	light->location.setValue(SbVec3f(10.0f, 10.0f, 0.0f));
	light->intensity.setValue(0.6f);
	_rootNode->addChild(light);


	// Flying wires
	SoSeparator* wireSep = new SoSeparator();

	// rotate the wires - Z axis
	SoElapsedTime* wiresZRotationTime = new SoElapsedTime();
	wiresZRotationTime->speed.setValue(0.08f);
	SoRotationXYZ* zRot = new SoRotationXYZ();
	zRot->axis.setValue(SoRotationXYZ::Z);
	zRot->angle.connectFrom(&(wiresZRotationTime->timeOut));
	wireSep->addChild(zRot);

	// rotate the wires - Y axis
	SoElapsedTime* wiresYRotationTime = new SoElapsedTime();
	wiresYRotationTime->speed.setValue(0.3f);
	SoRotationXYZ* yRot = new SoRotationXYZ();
	yRot->axis.setValue(SoRotationXYZ::Y);
	yRot->angle.connectFrom(&(wiresYRotationTime->timeOut));
	wireSep->addChild(yRot);

	// Create wires
	for (int i = 0; i < WIRES_COUNT; ++i) {
		// Generate random values for colour
		SbVec3f colour;
		for (int rgb = 0; rgb < 3; ++rgb) {
			colour[rgb] = std::rand() / (RAND_MAX+1.0);
		}

		// Generate random rotation
		SbVec3f rotation;
		for (int axis = 0; axis < 3; ++axis) {
			rotation[axis] = 2 * M_PI * (std::rand() / (RAND_MAX+1.0));
		}

		// Generate random translation
		float xTranslation = 2 * SB_SIZE_HALF * (std::rand() / (RAND_MAX+1.0)) - SB_SIZE_HALF;

		wireSep->addChild(createWire(SbColor(colour),
					SbVec3f(xTranslation, 0.0f, SB_SIZE_HALF/2),
					rotation[0], rotation[1], rotation[2]));
	}

	_rootNode->addChild(wireSep);
}

WireEnvironment::~WireEnvironment()
{
}

SoSeparator* WireEnvironment::createWire(const SbColor& colour,
		const SbVec3f& translation, float angleX, float angleY, float angleZ)
{
	SoSeparator* wireSep = new SoSeparator();

	// material
	SoMaterial* wireMaterial = new SoMaterial();
	wireMaterial->diffuseColor.setValue(colour);
	wireMaterial->specularColor.setValue(SbColor(1.0f, 1.0f, 1.0f));
	wireMaterial->shininess.setValue(0.8f);
	wireSep->addChild(wireMaterial);

	// translation
	SoTranslation* wireTranslation = new SoTranslation();
	wireTranslation->translation.setValue(translation);
	wireSep->addChild(wireTranslation);

	// rotate X axis
	SoRotationXYZ* xRot = new SoRotationXYZ();
	xRot->axis.setValue(SoRotationXYZ::X);
	xRot->angle.setValue(angleX);
	wireSep->addChild(xRot);

	// rotate Y axis
	SoRotationXYZ* yRot = new SoRotationXYZ();
	yRot->axis.setValue(SoRotationXYZ::Y);
	yRot->angle.setValue(angleY);
	wireSep->addChild(yRot);

	// rotate z axis
	SoRotationXYZ* zRot = new SoRotationXYZ();
	zRot->axis.setValue(SoRotationXYZ::Z);
	zRot->angle.setValue(angleZ);
	wireSep->addChild(zRot);

	// create the wire
	SoCylinder* wire1 = new SoCylinder();
	wire1->radius.setValue(0.5f);
	wire1->height.setValue(2 * WIRE_SIZE_HALF);
	wireSep->addChild(wire1);

	return wireSep;
}

void WireEnvironment::timeUpdate(float /* dt */)
{
	// do nothing
}

// Functions and constants needed by Factories::FactoryEnvironment
namespace
{
	// WireEnvironment identifier for the environments factory
	const std::string WIRE_ENVIRONMENT_ID = "Wire";

	// Function that returns an instance of the WireEnvironment class
	Environment *getWireEnvironmentInstance()
	{
		return new WireEnvironment::WireEnvironment();
	}

	// Register the selected environment at the environments factory
	const bool WIRE_ENVIRONMENT_REGISTERED = Factories::FactoryEnvironment::instance().
		registerObject(WIRE_ENVIRONMENT_ID, getWireEnvironmentInstance);
}
